/*
 2A0x.c
  N2A03/N2A07 CPU emulation, macro portion; #include's ops6502.h

 * MacFCEU 0.6.1
 *  revamped a lot of the code; strict variable usage, optimized
 *  no PUSH16/POP16
 *
 * Copyright notice for this file:
 *  x6502.c Copyright (C) 2002 Xodnizel
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <string.h>

#include "types.h"
#include "x6502.h"
#include "FCEU.h"
#include "debug.h"
#include "fsound.h"

#ifdef env68k
 #pragma segment N2A0x
#endif

X6502U		X6502F;
uint32		timestamp;

void FP_FASTAPASS(1) (*MapIRQHook)(int a);
void FP_FASTAPASS(1) (*DebugStepHook)(int c);

/* use these for reads, optimizing out zero extension */
#define RPC		X6502F.u32.PC
#define RA		X6502F.u32.A
#define RX		X6502F.u32.X

#define _A		XFrame.A
#define _X		XFrame.X
#define _Y		XFrame.Y
#define _S		XFrame.S
#define _P		XFrame.P
#define _PI		XFrame.mooPI
#define _DB		XFrame.DB
#define _DBlow		XFrame.DBlow
#define _count		XFrame.count
#define _tcount		XFrame.tcount
#define _IRQlow		XFrame.IRQlow
#define _jammed		XFrame.jammed

#define _ADDCYC(x)	(_tcount+=x,_count-=(x)*48,timestamp+=x)

/* I've managed to get all occurences of the read/write macros to use a variable, no expressions */
#define RMem(A)		(_DB=ARead[A](A))
#define _RMem16(A)	(_DBlow=ARead[A](A),_DB=ARead[(A+1)&0xFFFF]((A+1)&0xFFFF),X6502F.u32.DB)
#define WMem(A,V)	BWrite[A](A,V)

uint8 RMemNorm(unsigned A);
void WMemNorm(unsigned A,
	uint8 V);

#if defined(LOCALCODE) && !defined(env68k) /* all of these are always in Asm68K */
static uint8 INLINE RMemNorm(unsigned A)
{
	return RMem(A);
}

static void INLINE WMemNorm(unsigned A,
	uint8 V)
{
	WMem(A,V);
}

uint8 FASTAPASS(1) X6502_DMR(uint32 A)
{
	_ADDCYC(1);
	return RMem(A);
}

void FASTAPASS(2) X6502_DMW(uint32 A,
	uint8 V)
{
	_ADDCYC(1);
	WMem(A,V);
}

void FASTAPASS(1) X6502_IRQBegin(int w)
{
	XFrame.IRQlow|= w;
}

void FASTAPASS(1) X6502_IRQEnd(int w)
{
	XFrame.IRQlow&=~w;
}
#endif

#define PUSH(V)		((RAM+0x100)[_S--]=V)
#define POP()		(RAM+0x100)[++_S]

uint8		ZNTable[256];

/* parameters only used once, only _P and ZNTable are used during */
#define X_ZN(v)		_P&=~(Z_FLAG|N_FLAG),_P|=ZNT[v]
#define X_ZNT(v)	_P|=ZNT[v]

/* extra cycle in advent of page update (8 bit carry) on 6510 (650x doesn't... work) */

#define JR(cond)	do{if(!(cond)) ++_PC,_PC&=0xFFFF; else goto jr_destin;}while(0)
#define JR_DESTIN	jr_destin: {x=(int8)RMem(_PC),_ADDCYC(1), \
	t=++_PC,_PC+=x,_PC&=0xFFFF,t^=_PC; \
	if(t&=0x100) _ADDCYC(1); \
}

#define LDA		X_ZN(RA=x)
#define LDX		X_ZN(RX=x)
#define LDY		X_ZN(_Y=x)
#define LAX		X_ZN(RA=RX=x)
#define LAS		(x&=_S,_S=x,X_ZN(RA=RX=x))

/*  All of the freaky arithmetic operations. */
#define AND		X_ZN(RA&=x)
#define BIT		_P&=~(Z_FLAG|V_FLAG|N_FLAG),_P|=(!(x&RA))*Z_FLAG,_P|=x&(V_FLAG|N_FLAG)
#define EOR		X_ZN(RA^=x)
#define ORA		X_ZN(RA|=x)

/* N2A0x have no decimal mode at all. */

#define ADC		(t=_P&C_FLAG,t+=RA,t+=x,_P&=~(Z_FLAG|C_FLAG|N_FLAG|V_FLAG), \
	_P|=t>>8&C_FLAG,x^=RA,_P|=((t^RA)&~x)>>1&V_FLAG,x^=RA,X_ZNT(RA=t&=0xFF))
#define SBC		(x^=0xFF,ADC) //modifies x (don't use in SBC-extensions, like ISB)

#define CMPL(a1,a2)	(t=a1,t-=a2,_P&=~(Z_FLAG|N_FLAG|C_FLAG),_P|=~(t>>8)&C_FLAG,X_ZNT(t&=0xFF))

#define ANC		(t=RA&=x,_P&=~(Z_FLAG|N_FLAG|C_FLAG),_P|=t>>7,X_ZNT(t))

/* Special undocumented operation.  Very similar to CMP. */
#define AXS		(t=RA&RX,t-=x, \
	_P&=~(Z_FLAG|N_FLAG|C_FLAG),_P|=~(t>>8)&C_FLAG,X_ZNT(RX=t&=0xFF))

#define CMP		CMPL(RA,x)
#define CPX		CMPL(RX,x)
#define CPY		CMPL(_Y,x)

/* The following operations modify the byte being worked on. */
#define DEC		(--x,X_ZN(x&=0xFF))
#define INC		(++x,X_ZN(x&=0xFF))
#define DCP		(--x,x&=0xFF,CMP)
#define ISB		(x=(uint8)~(x+1),ADC,x^=0xFF)

#define ASL		(x+=x,_P&=~(C_FLAG|N_FLAG|Z_FLAG),_P|=x>>8,x&=0xFF,X_ZNT(x))
#define LSR		(_P&=~(C_FLAG|N_FLAG|Z_FLAG),_P|=x&C_FLAG,X_ZNT(x>>=1))

#define ROL		(x+=x,x+=(_P&C_FLAG),_P&=~(Z_FLAG|N_FLAG|C_FLAG),_P|=x>>8,X_ZNT(x&=0xFF))
#define ROR		(x+=(_P&C_FLAG)<<8,_P&=~(Z_FLAG|N_FLAG|C_FLAG),_P|=x&C_FLAG,X_ZNT(x>>=1))

#define ARR		(RA&=x,_P&=~V_FLAG,_P|=(_A^(_A>>1))&V_FLAG,t=RA>>7, \
	RA|=(_P&C_FLAG)<<8,_P&=~(C_FLAG|N_FLAG|Z_FLAG),_P|=t,X_ZNT(RA>>=1))
#define ASR		(RA&=x,_P&=~(C_FLAG|N_FLAG|Z_FLAG),_P|=_A&C_FLAG,X_ZNT(RA>>=1))

#define RLA		(x+=x,x+=_P&C_FLAG,_P&=~(Z_FLAG|N_FLAG|C_FLAG),_P|=x>>8,X_ZNT(RA&=x))
#define RRA		(x+=(_P&C_FLAG)<<8,_P&=~C_FLAG,_P|=x&C_FLAG,x>>=1,ADC)
#define SLO		(x+=x,_P&=~(C_FLAG|N_FLAG|Z_FLAG),_P|=x>>8,x&=0xFF,ORA)
#define SRE		(_P&=~(C_FLAG|N_FLAG|Z_FLAG),_P|=x&C_FLAG,X_ZNT(RA^=x>>=1))

#define SHA		(RA&RX&((A-_Y>>8)+1))
#define SHY		(_Y&((A-RX>>8)+1))
#define SHX		(RX&((A-_Y>>8)+1))
#define SHS		((_S=_A&_X)&((A-_Y>>8)+1))

/* XAA / ANE - BIG QUESTION MARK HERE */
/* bit #4 depends on what's left on bus - that means we can use whatever we want */
#define ANE		(RA|=0xEE,RA&=RX,AND)

/* Note that for GetABIRD and GetIYRD, FCE Ultra 0.98.13 did t^=0x100, not t-=0x100 as it should. */

/* Absolute */
#define GetAB(d)	(d=RMem(_PC),++_PC,_PC&=0xFFFF,d+=RMem(_PC)<<8,++_PC,_PC&=0xFFFF)

/* Absolute Indexed (for reads) */
#define GetABIRD(d,i)	{t=RMem(_PC)+i,++_PC,_PC&=0xFFFF,d=t+(RMem(_PC)<<8),++_PC,_PC&=0xFFFF; \
	if(t&=0x100) t=(d&=0xFFFF)-0x100,RMem(t),_ADDCYC(1);}

/* Absolute Indexed (for writes and rmws) */
#define GetABIWR(d,i)	(GetAB(t),d=t+i,d&=0xFFFF,t&=0xFF00,t|=d&0xFF,RMem(t))

#define GetIX(d)	(t=RMem(_PC),t+=RX,t&=0xFF,++_PC,_PC&=0xFFFF, \
	d=RRAM(t),++t,t&=0xFF,d+=RRAM(t)<<8)

/* Indirect Indexed (for reads) */
#define GetIYRD(d)	{d=RMem(_PC),++_PC,_PC&=0xFFFF,t=RRAM(d)+_Y,++d,d&=0xFF,d=t+(RRAM(d)<<8); \
	if(t&0x100) t=(d&=0xFFFF)-0x100,RMem(t),_ADDCYC(1);}

/* Indirect Indexed (for writes and rmws) */
#define GetIYWR(d)	(d=RMem(_PC),++_PC,_PC&=0xFFFF,t=RRAM(d)+_Y,++d,d&=0xFF,d=t+(RRAM(d)<<8), \
	t&=0x100,t=(d&=0xFFFF)-t,RMem(t))

/* Now come the macros to wrap up all of the above stuff addressing mode functions
   and operation macros.  Note that operation macros will always operate(redundant
   redundant) on the variable "x".
*/
#define RMW_A(op)	{x=RA; op; _A=x; break;} /* Meh... */
#define RMW_AB(op)	{GetAB(A),x=RMem(A),WMem(A,x); op; WMem(A,x); break;}
#define RMW_ABX(op)	{GetABIWR(A,RX); x=RMem(A),WMem(A,x); op; WMem(A,x); break;}
#define RMW_ABY(op)	{GetABIWR(A,_Y); x=RMem(A),WMem(A,x); op; WMem(A,x); break;}
#define RMW_IX(op)	{GetIX(A),x=RMem(A),WMem(A,x); op; WMem(A,x); break;}
#define RMW_IY(op)	{GetIYWR(A),x=RMem(A),WMem(A,x); op; WMem(A,x); break;}
#define RMW_ZP(op)	{A=RMem(_PC),x=RRAM(A),++_PC,_PC&=0xFFFF; op; WRAM(A,x); break;}
#define RMW_ZPX(op)	{A=(uint8)(RX+RMem(_PC)),x=RRAM(A),++_PC,_PC&=0xFFFF; op; WRAM(A,x); break;}

#define LD_IM(op)	{x=RMem(_PC),++_PC,_PC&=0xFFFF; op; break;}
#define LD_ZP(op)	{x=RMem(_PC),x=RRAM(x),++_PC,_PC&=0xFFFF; op; break;}
#define LD_ZPX(op)	{x=(uint8)(RX+RMem(_PC)),x=RRAM(x),++_PC,_PC&=0xFFFF; op; break;}
#define LD_ZPY(op)	{x=(uint8)(_Y+RMem(_PC)),x=RRAM(x),++_PC,_PC&=0xFFFF; op; break;}
#define LD_AB(op)	{GetAB(x),x=RMem(x); op; break;}
#define LD_ABX(op)	{GetABIRD(x,RX); x=RMem(x); op; break;}
#define LD_ABY(op)	{GetABIRD(x,_Y); x=RMem(x); op; break;}
#define LD_IX(op)	{GetIX(x),x=RMem(x); op; break;}
#define LD_IY(op)	{GetIYRD(x); x=RMem(x); op; break;}

#define ST_ZP(r)	{x=RMem(_PC),WRAM(x,r),++_PC,_PC&=0xFFFF; break;}
#define ST_ZPX(r)	{x=(uint8)(RX+RMem(_PC)),WRAM(x,r),++_PC,_PC&=0xFFFF; break;}
#define ST_ZPY(r)	{x=(uint8)(_Y+RMem(_PC)),WRAM(x,r),++_PC,_PC&=0xFFFF; break;}
#define ST_AB(r)	{GetAB(x),WMem(x,r); break;}
#define ST_ABX(r)	{GetABIWR(x,RX); WMem(x,r); break;}
#define ST_ABY(r)	{GetABIWR(x,_Y); WMem(x,r); break;}
#define ST_IX(r)	{GetIX(x),WMem(x,r); break;}
#define ST_IY(r)	{GetIYWR(x),WMem(x,r); break;}

uint8		CycTable[256] =
{ /* longest time: 8 cycles (line x3h) */
/*0x00*/ 7,6,2,8,3,3,5,5,3,2,2,2,4,4,6,6,
/*0x10*/ 2,5,2,8,4,4,6,6,2,4,2,7,4,4,7,7,
/*0x20*/ 6,6,2,8,3,3,5,5,4,2,2,2,4,4,6,6,
/*0x30*/ 2,5,2,8,4,4,6,6,2,4,2,7,4,4,7,7,
/*0x40*/ 6,6,2,8,3,3,5,5,3,2,2,2,3,4,6,6,
/*0x50*/ 2,5,2,8,4,4,6,6,2,4,2,7,4,4,7,7,
/*0x60*/ 6,6,2,8,3,3,5,5,4,2,2,2,5,4,6,6,
/*0x70*/ 2,5,2,8,4,4,6,6,2,4,2,7,4,4,7,7,
/*0x80*/ 2,6,2,6,3,3,3,3,2,2,2,2,4,4,4,4,
/*0x90*/ 2,6,2,6,4,4,4,4,2,5,2,5,5,5,5,5,
/*0xA0*/ 2,6,2,6,3,3,3,3,2,2,2,2,4,4,4,4,
/*0xB0*/ 2,5,2,5,4,4,4,4,2,4,2,4,4,4,4,4,
/*0xC0*/ 2,6,2,8,3,3,5,5,2,2,2,2,4,4,6,6,
/*0xD0*/ 2,5,2,8,4,4,6,6,2,4,2,7,4,4,7,7,
/*0xE0*/ 2,6,3,8,3,3,5,5,2,2,2,2,4,4,6,6,
/*0xF0*/ 2,5,2,8,4,4,6,6,2,4,2,7,4,4,7,7,
};

/* remember that "reset" does not initialize like "power" */
/* at least one cart takes advantage of this fact (press reset thirteen times) */
void X6502_Reset(void)
{
	XFrame.IRQlow = FCEU_IQRESET;
}

void X6502_Power(void)
{
	memset((uint32*)&XFrame,0,sizeof(XFrame));
	X6502_Reset();
}

void X6502_Init(void)
{ /*no more external calls, unaligned, small*/
	memset(ZNTable,0,0x80);
	memset(0x80+ZNTable,N_FLAG,0x80);
	ZNTable[0] = Z_FLAG;
}

#ifdef env68k
 #undef RMem
 #undef WMem
 #define RMem		RMemNorm
 #define WMem		WMemNorm
#endif

/* reused - optimization purposes */
#define t		cycles

#if (defined(C80x86) && defined(__GNUC__))
 #define EDI		asm ("edi") /* Gives a nice little speed boost. */
#else
 #define EDI
#endif

#define _SETALLP(x)	(*(uint16*)&_PI=x*0x0101)

#define X6502_RUN_FUNC	\
	register const uint8	*ZNT=(const uint8*)ZNTable; \
	register int32 A; \
	register uint32	_PC EDI=RPC; \
	register uint32	x=cycles<<4; \
	if (PAL) /* *=15 */ \
		x-=cycles; \
	_count+= x; \
	while(_count>0) { \
		if (_IRQlow!=0) { \
			if (_IRQlow&FCEU_IQRESET) { \
				_PC = _RMem16(0xFFFC); \
				_jammed = 0; \
				_SETALLP(I_FLAG); \
				_IRQlow&=~FCEU_IQRESET; \
			} else if (_IRQlow&FCEU_IQNMI2) { \
				_IRQlow&=~FCEU_IQNMI2; \
				_IRQlow|=FCEU_IQNMI; \
			} else if (!_jammed) \
				if (_IRQlow&FCEU_IQNMI) { \
					_ADDCYC(7); \
					PUSH(_PC>>8); PUSH(_PC); \
					PUSH((_P&~B_FLAG)|U_FLAG); \
					_P|=I_FLAG; \
					_PC=_RMem16(0xFFFA); \
					_IRQlow&=~FCEU_IQNMI; \
				} else if ((_PI&I_FLAG)==0) { \
					_ADDCYC(7); \
					PUSH(_PC>>8); PUSH(_PC); \
					PUSH((_P&~B_FLAG)|U_FLAG); \
					_P|=I_FLAG; \
					_PC=_RMem16(0xFFFE); \
				} \
			_IRQlow&=~FCEU_IQTEMP; \
			if (_count<=0) { \
				_PI=_P; \
				goto end; \
			} \
		} \
		if (DebugStepHook) { \
			XFrame.PC = _PC; \
			DebugStepHook(0); \
			_PC = RPC; \
		} \
		_PI=_P; \
		t=RMem(_PC),x=CycTable[t],_ADDCYC(x); \
		x=_tcount, _tcount=0; \
		if (MapIRQHook) \
			MapIRQHook(x); \
		FCEU_SoundCPUHook(x);

void FP_FASTAPASS(1) X6502_RunFast(int32);
void FP_FASTAPASS(1) X6502_RunFull(int32);

/* if not local, then assembly will pick it up */
/* it takes over ten minutes to compile with optimizations on using SC! */
#ifdef LOCALCODE
/* to get all gg/cheats to work, treat RAM as Mem... it is slower */
#define RRAM		RMem
#define WRAM		WMem

void FP_FASTAPASS(1) X6502_RunFull(register int32 cycles)
{
	X6502_RUN_FUNC
	exec:
		++_PC, _PC&=0xFFFF;
		switch(t)
		{
		#include "ops6502.h"
		}
	}
end:
	XFrame.PC = _PC;
}

/* only used by Zero Page modes */
#undef RRAM
#undef WRAM
#define RRAM(A)		RAM[A]
#define WRAM(A,d)	(RAM[A]=d)

void FP_FASTAPASS(1) X6502_RunFast(register int32 cycles)
{
	X6502_RUN_FUNC
	exec:
		++_PC, _PC&=0xFFFF;
		switch(t)
		{
		#include "ops6502.h"
		}
	}
end:
	XFrame.PC = _PC;
}
#endif

void FP_FASTAPASS(1) (*X6502_Run)(int32 cycles) = X6502_RunFast;

void X6502_DebugLevel(int bits)
{
	static int	level;

	if (bits>0)
		level|= bits;
	else
		level&= bits;
	X6502_Run = level ? X6502_RunFull:X6502_RunFast;
}
