/*
 WAVE.c
  WAVE file recording

 * MacFCEU 0.6a4
 *
 * See AUTHORs for original credits
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <stdio.h>

#include "driver.h"
#include "FCEU.h"
#include "fsound.h"
#include "WAVE.h"

#ifdef env68k
 #pragma segment files
#endif

static FILE	*soundlog = 0;
static long	wsize;
static uint8	toupdate = 1;
static uint8	current_wavedepth;
int		next_wavedepth = 16;

void FCEU_WaveUpdateHeader(void)
{
	long	s, o;

	if (soundlog==NULL || !toupdate || (o=ftell(soundlog))<0
			|| fseek(soundlog,4,SEEK_SET)<0)
		return;
	fputc((o-=8)&0xFF,soundlog);
	fputc(o>>8&0xFF,soundlog);
	fputc(o>>16&0xFF,soundlog);
	fputc(o>>24&0xFF,soundlog), o += 8;

	fseek(soundlog,0x28,SEEK_SET);
	s = wsize;
	fputc(s&0xFF,soundlog);
	fputc(s>>8&0xFF,soundlog);
	fputc(s>>16&0xFF,soundlog);
	fputc(s>>24&0xFF,soundlog);

	fseek(soundlog,o,SEEK_SET);

	toupdate = 0;
}

/* Checking whether the file exists before wiping it out is left up to the 
   reader..err...I mean, the driver code, if it feels so inclined(I don't feel
   so).
*/
void FCEU_WriteWaveData(int16 *Buffer,
	int Count)
{
#ifdef mac
	static int16	temp[8192]; /* default maximum */
#else
	int16	temp[Count];	/* Yay.  Is this the first use of this "feature" of C in FCE Ultra? */
#endif
	int16	*dest;
	int	x;

#ifdef mac
	if (Count>8192)
	{
		/*DebugStr("\pFCEU_WriteWaveData: temporary buffer overflow");*/
		Count = 8192;
	}
#endif
	if (soundlog==NULL || (x=Count)<=0)
		return;

	dest = temp;

	if (current_wavedepth==8)
		do{*((uint8*)dest)++ = (*Buffer++>>8)^0x80;} while(--x!=0);
	else /* otherwise assume 16 */
		do
		{
		#if BIG_ENDIAN
			((uint8*)dest)[1] = *((uint8*)Buffer)++;
			((uint8*)dest)[0] = *((uint8*)Buffer)++;
		#else
			*dest = *Buffer++;
		#endif
		} while(++dest,--x!=0);

	wsize+= fwrite(temp,1,(uint8*)dest-(uint8*)temp,soundlog);

	toupdate = 1;
}

int FCEUI_EndWaveRecord(void)
{
	if (soundlog==NULL)
		return 0;

	FCEU_WaveUpdateHeader();

	fclose(soundlog);
	soundlog = 0;

	return 1;
}


int FCEUI_BeginWaveRecord(char *fn)
{
	int		r;

	if ((soundlog=FCEUD_UTF8fopen(fn,"wb"))==NULL)
		return 0;
	wsize = 0;


/* Write the header. */
	fwrite("RIFF\0\0\0\0WAVEfmt ",4+4+8,1,soundlog);

	fputc(0x10,soundlog);
	fputc(0,soundlog);
	fputc(0,soundlog);
	fputc(0,soundlog);

	fputc(1,soundlog);	// PCM
	fputc(0,soundlog);

	fputc(1,soundlog);	// Monophonic
	fputc(0,soundlog);

	r = FSettings.SndRate;
	fputc(r&0xFF,soundlog);
	fputc((r>>8)&0xFF,soundlog);
	fputc((r>>16)&0xFF,soundlog);
	fputc((r>>24)&0xFF,soundlog);

	if (next_wavedepth==8)
		current_wavedepth = 8;
	else
		current_wavedepth = 16;

	r<<= current_wavedepth>>4;
	fputc(r&0xFF,soundlog);
	fputc((r>>8)&0xFF,soundlog);
	fputc((r>>16)&0xFF,soundlog);
	fputc((r>>24)&0xFF,soundlog);
	fputc(current_wavedepth>>3,soundlog);
	fputc(0,soundlog);
	fputc(current_wavedepth,soundlog);
	fputc(0,soundlog);

	fputs("data",soundlog);
	fseek(soundlog,4,SEEK_CUR);

	return (toupdate=1);
}
