/*
 boards/latch.c
  mappers 2(UxROM), 3(CNROM), 7(AxROM), 11/144(AGCI), 13(CPROM), 38,
	59(T3H53), 66(MHROM/GNROM), 70, 78(JF16), 79(AVE NINA06),
	93(Sunsoft-2A), 94(UN1ROM), 107, 140(JF14), 152, 240(C&E SHLZ)
  boards UNL-CC21

 * MacFCEU 0.6.1
 *
 * Copyright notice for this file:
 *  59.c, 79.c, datalatch.c (C) 2002 Xodnizel
 *  addrlatch.cpp (C) 2006 CaH4e3
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "mapinc.h"
#include "latch.h"

#define CHRRAM (GameMemBlock)

uint8		latch_byte;
uint16		latch_addr;

static uint32	latch_init;
static uint32	addrlohi;
static void	(*WSync)(void);
static readfunc	rfunc;

DECLFW(Latch_BusWrite)
{
	latch_addr = A;
	latch_byte = ARead[A](A);
	//should be able to get how far into the instruction the CPU is
	if (V!=latch_byte && (XFrame.count>0 || int64lo(timestampbase)&2))
		latch_byte = V;
	WSync();
}

DECLFW(Latch_Write)
{
	latch_addr = A;
	latch_byte = V;
	WSync();
}

static void LatchRestore(int v)
{
(void)v;
	WSync();
}

/*		flags	prgshr	prgmsk	chrshr	chrmsk	mirxsm
UNL-CC21	$1E	-	-	0	1	1
BMC-190in1	$3D	2	7	2	7	~0
ANROM		$16	0	15	-	-	4
GNROM		$10	4	3	0	3	-
CNROM		$30	+1	-	0	*	-
CPROM		$20	+1	-	-/0	0	-
MHROM		$10	4	-	0	15	-
UNROM		$30	0,-1	-	-	-	-
BMCA65AS	$54/$56	1	15/special	0	special
Mapper11	$10	0	15	4	-	-
Mapper13	$00	0	0	0	3	-
Mapper36	$10	4	-	0	15	-
Mapper38	$10	0	3	2	-	-
Mapper58	$5D	+1	7	3	7	~7
Mapper70	$30	4,-1	-	0	15	-
Mapper78	$36	0,-1	7	4	-	3
Mapper87	$30	+1	0	(special)	-
Mapper94	$30	2,-1	-	-	-	-
Mapper101	$30	+1	0	0	-	-
Mapper107	$10	1	3	0	7	-
Mapper113	$10	3	7	(special)	(HES 6in1)
Mapper144	(Mapper11, different addressing)
Mapper152	$36	4,-1	7	0	15	7
Mapper180	$30	0,-2	-	-	-	-
Mapper184	$20	-,+1	-	0/4	-	-
Mapper200	$3C	0	7	0	7	3
Mapper240	(MHROM, different addr. range)
*/

#if 0
/* [pre.s2] [ps1.3] [ps2.3] [pmk.8] [mie.3] [cs1.3] [cs2.3] [cmk.7] */
static uint32	prchsm;

#define MAKEPCSM(pre,ps1,ps2,pmk,mie,cs1,cs2,cmk) \
	((((((((pre<<3)+ps1<<3)+ps2<<8)+(pmk&255)<<3)+mie<<3)+cs1<<3)+cs2<<7)+(cmk&127))

static void VSync(void)
{
	uint8		b, sm;
	uint8		flag = latch_init>>16;

	sm = prchsm>>16;
	b = latch_byte>>(sm>>11&7), b&= sm&0xFF;
	if ((latch_byte&flag&PSELMASK)|(flag&FORCEP16))
	{
		if ((sm=(int16)sm>>14)==0)
			sm = 1;
		sm = b;
		setprg16(0x8000,b);
		setprg16(0xC000,sm);
	} else
		setprg32(0x8000,b>>(sm>>8&7));

	sm = prchsm;
	b = latch_byte>>(sm>>7&7), b&= sm&0x7F;
	if (flag&CHR8OVR4)
		setchr8(b);
	else
	{
		setchr4(0,b);
		b = latch_byte>>(sm>>10&7);
		if ((sm&0xFF)==0)
			b&= 3;
		setchr4(0x1000,b);
	}

	if (flag&DOMIRROR)
		setmirror((flag&MIRFOURS)+((latch_byte>>(sm>>13)&1)^(flag&MIRINVHV)));
}

static void ASync(void)
{
	uint16		b, sm;
	uint8		flag = latch_init>>16;

	sm = prchsm>>16;
	b = latch_addr>>(sm>>11&7), b&= sm&0xFF;
	if ((latch_addr&flag&PSELMASK)|(flag&FORCEP16))
	{
		if ((sm=(int16)sm>>14)==0)
			sm = 1;
		if (sm==-2)
			sm = b, b = 0;
		setprg16(0x8000,b);
		setprg16(0xC000,sm);
	} else
		setprg32(0x8000,b>>(sm>>8&7));

	sm = prchsm;
	b = latch_addr>>(sm>>7&7), b&= sm&0x7F;
	if (flag&CHR8OVR4)
		setchr8(b);
	else
	{
		setchr4(0,b);
		b = latch_addr>>(sm>>10&7);
		if ((sm&0xFF)==0)
			b&= 3;
		setchr4(0x1000,b);
	}

	if (flag&DOMIRROR)
		setmirror((flag&MIRFOURS)+((latch_addr>>(sm>>13)&1)^(flag&MIRINVHV)));
}

static void Latch_Close(void)
{
	prchsm = 0;
}
#endif //#if 0


static void M11Sync(void); //as default

void Latch_Power(void)
{
	writefunc	wf = latch_init&0x800000U ? Latch_BusWrite:Latch_Write;

	wf(latch_init&0xFFFF,latch_init>>24);
	SetWriteHandler(addrlohi>>16,addrlohi&0xFFFF,wf);
	SetReadHandler(0x8000,0xFFFF,rfunc);
}

void Latch_Init(CartInfo *info,
	void (*wsyncproc)(void),
	readfunc func,
	uint32 init,
	uint32 lohi)
{
	latch_init = init;
	addrlohi = lohi;

	if ((WSync=wsyncproc)==0)
		WSync = M11Sync;
	AddExState(&latch_byte,1,0,"LATC");

	if (init&0xFFFF)
	{
		AddExState(&latch_addr,2,1,"LATA");
//		if (wsyncproc==0)
//			WSync = ASync;
	}

	info->Power = Latch_Power;
//	info->Close = Latch_Close;

	if (func)
		rfunc = func;
	else
		rfunc = CartBR;
	GameStateRestore = LatchRestore;
}

static void CapHIPower(void)
{
	setprg16(0xC000,~0);
	Latch_Power();
}

/* ---- mapper 2: UxROM */

static void UNROMSync(void)
{
	setprg16(0x8000,latch_byte); //ROM_BANK16
	setchr8(0);
}

/*static void UxROMPower(void)
{
//	setvram8(CHRRAM);
}*/

void UxROM_Init(CartInfo *info)
{
	Latch_Init(info,UNROMSync,0,0,0x8000FFFFU);
	info->Power = CapHIPower;
//	if (ROM_size<256/4) PRGmask16[0]&= 7;
//	AddExState(CHRRAM,8192,0,"CHRR");
}

/* ---- mapper 3: CNROM */

static void CNROMSync(void)
{
	setchr8(latch_byte); //&=3
	setprg16(0x8000,0);
	setprg16(0xC000,1);
}

void CNROM_Init(CartInfo *info)
{
	Latch_Init(info,CNROMSync,0,0,0x8000FFFFU);
}

/* ---- mapper 7: ANROM/AOROM */

static void AxROMSync(void)
{
	setprg32(0x8000,latch_byte&15);
	setmirror(MI_0+(latch_byte>>4&1));
}

void AxROM_Init(CartInfo *info)
{
	Latch_Init(info,AxROMSync,0,0,0x8000FFFFU);
}

/* ---- mapper 11 (AGCI-47516/ColorDreams 74*377), 144 (AGCI 50282) */

static void M11Sync(void)
{
	setprg32(0x8000,latch_byte&15);
	setchr8(latch_byte>>4);
}

void AGCI47516_Init(CartInfo *info)
{
	Latch_Init(info,0,0,0,0x8000FFFFU);
}

void AGCI50282_Init(CartInfo *info)
{
	Latch_Init(info,0,0,0,0x8001FFFFU);
}

/* ---- mapper 13: 32KB VRAM CPROM */

static void CPROMSync(void)
{
//	setvram4(0x1000,CHRRAM+((latch_byte&=3)<<12));
	setchr4r(0x10,0x1000,latch_byte&3);
	setprg32(0x8000,latch_byte>>4&3);
}

/*static void CPROMPower(void)
{
//	setvram8(0);
	Latch_Power();
}
*/

void CPROM_Init(CartInfo *info)
{
	Latch_Init(info,CPROMSync,0,0,0x8000FFFFU);
//	info->Power = CPROMPower;
	AddExState(CHRRAM,16384,0,"CHRR");
	SetupCartCHRMapping(0x10,CHRRAM,16384,1);
}

/* ---- mapper 34: BxROM or AVE NINA-001 */

static void BNROMSync(void)
{
	setprg32(0x8000,latch_byte);
	setchr8r(0x10,0);
}

void BNROM_Init(CartInfo *info)
{
	Latch_Init(info,BNROMSync,0,0,0x8000FFFF);
	AddExState(CHRRAM,8192,0,"CHRR");
	SetupCartCHRMapping(0x10,CHRRAM,8192,1);
}

/*void BxROM_Init(CartInfo *info)
{
	if (VROM_size)
		{AVE_NINA001_Init(info); return;}
	BNROM_Init(info);
}*/

/* ---- mapper 38: 74*161/138 */

static void M38Sync(void)
{
	setprg32(0x8000,latch_byte&3);
	setchr8(latch_byte>>2);
}

void Discrete_74_161_138_Init(CartInfo *info)
{
	Latch_Init(info,M38Sync,0,0,0x70007FFFU);
}

/* ---- mapper 66, 140/240: GNROM compatible, MHROM */

/*static void M66Sync2(void)
{
	setprg32(latch_byte>>4);
	VRAM_BANK8(latch_byte&3);
}
*/

static void GNROMSync(void)
{
//	latch_byte&= 0x33;
	setprg32(0x8000,latch_byte>>4);
	setchr8(latch_byte);
}

void GNROM_Init(CartInfo *info)
{ /* notey: this can have bus conflicts */
	Latch_Init(info,GNROMSync,0,0,0x8000FFFFU);
}

void MHROM_Init(CartInfo *info)
{
	GNROM_Init(info);
//	PRGmask32[0]&= 1;
//	CHRmask8[0]&= 1;
}

void JF14x_Init(CartInfo *info)
{
	Latch_Init(info,GNROMSync,0,0,0x60007FFFU);
}

void CandE_SHLZ_Init(CartInfo *info)
{
	Latch_Init(info,GNROMSync,0,0,0x40205FFFU);
}

/* ---- mapper 70(type A), 152(type B): 74*161/161/32 */

static void M70Sync(void)
{
	setprg16(0x8000,latch_byte>>4);
	setchr8(latch_byte&15);
}

void Discrete_74_161_161_32A_Init(CartInfo *info)
{
	Latch_Init(info,M70Sync,0,0,0x8000FFFFU);
	info->Power = CapHIPower;
}

static void M152Sync(void)
{
	setprg16(0x8000,latch_byte>>4&7);
	setchr8(latch_byte&15);
	setmirror(MI_0+(latch_byte>>7&1));
}

void Discrete_74_161_161_32B_Init(CartInfo *info)
{
	if (info->mirror<=MI_V)
		{Discrete_74_161_161_32A_Init(info); return;}
	Latch_Init(info,M152Sync,0,0,0x8000FFFFU);
	info->Power = CapHIPower;
}

/* ---- mapper 78: Jaleco JF-16 */

static void M78Sync(void)
{
	setprg16(0x8000,latch_byte&7);
	setchr8(latch_byte>>4);
	setmirror(MI_0+(latch_byte>>3&1));
}

void JF16_Init(CartInfo *info)
{
	Latch_Init(info,M78Sync,0,0,0x8000FFFFU);
	info->Power = CapHIPower;
}

/* ---- mapper 79: AVE NINA-06 */

static void M79Sync(void)
{
	setprg32(0x8000,latch_byte>>3);
	setchr8(latch_byte);
}

static void M79Power(void)
{
	int		i;

	Latch_Power();
	for(i=0x4300; i<0x6000; i+=0x200)
		SetWriteHandler(i,i+0xFF,Latch_Write);
}

void AVE_NINA06_Init(CartInfo *info)
{
	Latch_Init(info,M79Sync,0,0,0x410041FF);
	info->Power = M79Power;
}

/* ---- mapper 87: 74*139/74 */

static void M87Sync(void)
{
	setprg16(0x8000,0);
	setprg16(0xC000,1);
	setchr8((latch_byte>>1&1)|(latch_byte<<1&2));
}

void Discrete_74_139_74_Init(CartInfo *info)
{
	Latch_Init(info,M87Sync,0,0xFFU<<24,0x6000FFFFU);
}

/* ---- mapper 93: Sunsoft-2A - "Shanghi" original version mapper */

static void Sunsoft2ASync(void)
{
	setprg16(0x8000,latch_byte>>4);
	setchr8(0);
}

void Sunsoft2A_Init(CartInfo *info)
{
	Latch_Init(info,Sunsoft2ASync,0,0,0x8000FFFFU);
	info->Power = CapHIPower;
}

/* ---- mapper 94: UN1ROM */

static void UN1ROMSync(void)
{
	setprg16(0x8000,latch_byte>>2);
	setchr8(0);
}

void UN1ROM_Init(CartInfo *info)
{
	Latch_Init(info,UN1ROMSync,0,0,0x8000FFFFU);
	info->Power = CapHIPower;
}

/* ---- mapper 107: Magic Series */

static void M107Sync(void)
{
	setprg32(0x8000,latch_byte>>1&3);
	setchr8(latch_byte&7);
}

void Magic_Series_Init(CartInfo *info)
{
	Latch_Init(info,M107Sync,0,0xFFU<<24,0x8000FFFFU);
}

/* ---- mapper 59: T3H53 */

static void M59Sync(void)
{
	unsigned	A = latch_addr;

	setprg32(0x8000,A>>4&7);
	setchr8(A&7);
	setmirror(A>>3&1); //MIRROR_SET2
}

static DECLFR(M59R8000)
{
	if (latch_addr&0x100)
		return 0;
	else
		return CartBR(A);
}

void T3H53_Init(CartInfo *info)
{
	Latch_Init(info,M59Sync,M59R8000,0x8000,0x8000FFFFU);
}
