/*
 cart.c

 * MacFCEU 0.6.1
 *
 * Copyright notice for this file:
 *  cart.c (C) 2002 Xodnizel
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <string.h>
#include <stdlib.h>
#include <stdio.h>

#include "driver.h"
#include "FCEU.h"
#include "PPU.h"

#include "cart.h"
#include "memory.h"

#include "general.h"

/*
   This file contains all code for coordinating the mapping in of the
 address space external to the NES.
   It's also (ab)used by the NSF code.
*/

uint32	ROM_size, VROM_size; /* totals (prev. in ines.c); *16KB PRG, *8KB CHR */

uint8	*Page[32], *VPage[8];
uint8	**VPageR = VPage;
uint8	*MMC5SPRVPage[8];
uint8	*MMC5BGVPage[8];

uint8	PRGIsRAM[32]; /* This page is/is not PRG RAM. */

/* 16 are (sort of) reserved for UNIF/iNES and 16 to map other stuff. */
static uint8	CHRram[32];
static uint8	PRGram[32];

uint8	*PRGptr[32];
uint8	*CHRptr[32];

uint32	PRGsize[32];
uint32	CHRsize[32];

uint32	PRGmask2[32];
uint32	PRGmask4[32];
uint32	PRGmask8[32];
uint32	PRGmask16[32];
uint32	PRGmask32[32];

uint32	CHRmask1[32];
uint32	CHRmask2[32];
uint32	CHRmask4[32];
uint32	CHRmask8[32];

/*static INLINE void setpageptr(int s, uint32 A, uint8 *p, int ram)*/
#define setpageptr(s,A,p,ram) \
	do{uint8 *_p; uint32 _AB=(A)>>11; int _x,_r=ram; \
	if((s)>>1>0) \
		if(_AB+=(_x=((s)>>1))-1,(_p=p)==0) \
			do{ \
				PRGIsRAM[_AB]=0; \
				Page[_AB]=0; \
			} while(--_AB,--_x!=0); \
		else { \
			_p-=A; \
			do{ \
				PRGIsRAM[_AB]=_r; \
				Page[_AB]=_p; \
			} while(--_AB,--_x!=0); \
		} \
} while(0)

static uint8	nothing[8192];

void ResetCartMapping(void)
{
	int		x;
//2010 bug fix: I put this as --x!=0 fm 32/7 (writing at the wrong addresses)
	x = 31;
	do
	{
		Page[x] = nothing-(x<<11);
		PRGptr[x] = CHRptr[x] = 0;
		PRGsize[x] = CHRsize[x] = 0;
	} while(--x>=0);
	x = 7;
	do
		MMC5SPRVPage[x] = MMC5BGVPage[x] = VPageR[x] = nothing-(x<<10);
	while(--x>=0);
}

void SetupCartPRGMapping(int chip,
	uint8 *p,
	uint32 size,
	int ram)
{
	PRGptr[chip] = p;
	PRGsize[chip] = size;

	PRGmask2[chip] = (size>>11)-1;
	PRGmask4[chip] = (size>>12)-1;
	PRGmask8[chip] = (size>>13)-1;
	PRGmask16[chip] = (size>>14)-1;
	PRGmask32[chip] = (size>>15)-1;

	PRGram[chip] = ram ? 1:0;
}

void SetupCartCHRMapping(int chip,
	uint8 *p,
	uint32 size,
	int ram)
{
	CHRptr[chip] = p;
	CHRsize[chip] = size;

	CHRmask1[chip] = (size>>10)-1;
	CHRmask2[chip] = (size>>11)-1;
	CHRmask4[chip] = (size>>12)-1;
	CHRmask8[chip] = (size>>13)-1;

	CHRram[chip] = ram;
}

//added these 3; 2010.03.12
int NewCHRRAM(int w,
	uint32 z)
{
	uint8		*r;

	if ((r=(uint8*)FCEU_gmalloc(z))==0)
		return 0;
	SetupCartCHRMapping(w,r,z,1);

	return 1;
}

void DeleteCHRRAM(int w)
{
	if (CHRptr[w] && CHRram[w])
		{FCEU_free(CHRptr[w]); CHRptr[w] = nothing, CHRram[w] = 0;}
}

void EraseCHRRAM(int w,
	int32 v)
{
	if (CHRptr[w] && CHRram[w])
		FCEU_memsetl(CHRptr[w],v,CHRsize[w]);
}

DECLFR(CartBR)
{
	return Page[A>>11][A];
}

DECLFW(CartBW)
{
	uint32		AB = A>>11;

	if (PRGIsRAM[AB] && Page[AB])
		Page[AB][A] = V;
}

DECLFR(CartBROB)
{
	uint32		AB = A>>11;

	if (!Page[AB])
		return A>>8; //XFrame.DB
	return Page[AB][A];
}

void FASTAPASS(3) setprg2r(int r,
	unsigned int A,
	unsigned int V)
{
	V&= PRGmask2[r];
	setpageptr(2,A,PRGptr[r]==0 ? 0:&PRGptr[r][V<<11],PRGram[r]);
}

void FASTAPASS(3) setprg4r(int r,
	unsigned int A,
	unsigned int V)
{
	V&= PRGmask4[r];
	setpageptr(4,A,PRGptr[r]==0 ? 0:&PRGptr[r][V<<12],PRGram[r]);
}

void FASTAPASS(3) setprg8r(int r,
	unsigned int A,
	unsigned int V)
{
	uint8		*p = PRGptr[r];

	if (PRGsize[r]>=1<<13)
	{
		V&= PRGmask8[r];
		setpageptr(8,A,p==0 ? 0:&p[V<<13],PRGram[r]);
	} else
	{
		uint32		VA = V<<2;
		int		x = 0, m = PRGmask2[r];

		do
			setpageptr(2,A,p==0 ? 0:&p[((VA+x)&m)<<11],PRGram[r]);
		while(A+=1<<11,++x<4);
	}
}

void FASTAPASS(3) setprg16r(int r, unsigned int A, unsigned int V)
{
	uint8		*p = PRGptr[r];

	if (PRGsize[r]>=1<<14)
	{
		V&= PRGmask16[r];
		setpageptr(16,A,p==0 ? 0:&p[V<<14],PRGram[r]);
	} else
	{
		uint32		VA = V<<3;
		int		x = 0, m = PRGmask2[r];

		do
			setpageptr(2,A,p==0 ? 0:&p[((VA+x)&m)<<11],PRGram[r]);
		while(A+=1<<11,++x<8);
	}
}

void FASTAPASS(3) setprg32r(int r,
	unsigned int A,
	unsigned int V)
{
	uint8		*p = PRGptr[r];

	if (PRGsize[r]>=1<<15)
	{
		V&= PRGmask32[r];
		setpageptr(32,A,p==0 ? 0:&p[V<<15],PRGram[r]);
	} else
	{
		uint32		VA = V<<4;
		int		x = 0, m = PRGmask2[r];

		do
			setpageptr(2,A,p==0 ? 0:&p[((VA+x)&m)<<11],PRGram[r]);
		while(A+=1<<11,++x<16);
	}
}

void FASTAPASS(3) setchr1r(int r,
	unsigned int A,
	unsigned int V)
{
	uint32		AB;

	if (CHRptr[r]==0)
		return;
	FCEUPPU_LineUpdate();

	AB = A>>10, V&= CHRmask1[r];
	VPageR[AB] = &CHRptr[r][V<<10]-A;

	if (CHRram[r])
		PPUCHRRAM|= 1<<AB;
	else
		PPUCHRRAM&=~(1<<AB);
}

void FASTAPASS(3) setchr2r(int r,
	unsigned int A,
	unsigned int V)
{
	uint32		AB;

	if (CHRptr[r]==0)
		return;
	FCEUPPU_LineUpdate();

	AB = A>>10, V&= CHRmask2[r];
	VPageR[AB] = VPageR[AB+1] = &CHRptr[r][V<<11]-(A);

	if (CHRram[r])
		PPUCHRRAM|= 3<<AB;
	else
		PPUCHRRAM&=~3<<AB;
}

void FASTAPASS(3) setchr4r(int r,
	unsigned int A,
	unsigned int V)
{
	uint32		AB;

	if (CHRptr[r]==0)
		return;
	FCEUPPU_LineUpdate();

	AB = A>>10, V&= CHRmask4[r];
	VPageR[AB] = VPageR[AB+1] =
		VPageR[AB+2] = VPageR[AB+3] = &CHRptr[r][V<<12]-(A);

	if (CHRram[r])
		PPUCHRRAM|= 15<<AB;
	else
		PPUCHRRAM&=~(15<<AB);
}

void FASTAPASS(2) setchr8r(int r,
	unsigned int V)
{
	if (CHRptr[r]==0)
		return;

	FCEUPPU_LineUpdate();

	V&= CHRmask8[r], V<<= 13;
	setvram8(&CHRptr[r][V]);

	if (CHRram[r])
		PPUCHRRAM|= 0xFF;
	else
		PPUCHRRAM&=~0xFF;
}

void FASTAPASS(1) setvram8(uint8 *p)
{
	register uint8	**w = VPageR;
	/* unrolled */
	*w++ = p, *w++ = p, *w++ = p, *w++ = p,
	*w++ = p, *w++ = p, *w++ = p, *w = p;
	PPUCHRRAM|= 0xFF;
}

void FASTAPASS(2) setvram4(uint32 A, uint8 *p)
{
	uint32		AB = A>>10;
	register uint8	**w = AB+VPageR;
	/* unrolled */
	p-= A;
	*w++ = p, *w++ = p, *w++ = p, *w = p;
	PPUCHRRAM|= 15<<AB;
}

void FASTAPASS(3) setvramb1(uint8 *p, uint32 A, uint32 b)
{
	uint32		AB;

	FCEUPPU_LineUpdate();

	AB = A>>10;
	VPageR[AB] = p+(b<<10)-A;
	PPUCHRRAM|= 1<<AB;
}

void FASTAPASS(3) setvramb2(uint8 *p, uint32 A, uint32 b)
{
	uint8		**w;
	uint32		AB;

	FCEUPPU_LineUpdate();

	AB = A>>10, w = AB+VPageR;
	w[0] = w[1] = p+(b<<11)-A;
	PPUCHRRAM|= 3<<AB;
}

void FASTAPASS(3) setvramb4(uint8 *p, uint32 A, uint32 b)
{
	uint8		**w;
	uint32		AB;

	FCEUPPU_LineUpdate();

	AB = A>>10, w = AB+VPageR;
	w[0] = w[1] = w[2] = w[3] = p+(b<<12)-A;
	PPUCHRRAM|= 15<<AB;
}

void FASTAPASS(2) setvramb8(uint8 *p, uint32 b)
{
	uint8		**w;

	FCEUPPU_LineUpdate();

	w = VPageR;
	w[0] = w[1] = w[2] = w[3] = w[4] = w[5] = w[6] = w[7] = p+(b<<13);

	PPUCHRRAM|= 0xFF;
}

/* This function can be called without calling SetupCartMirroring(). */

void FASTAPASS(3) setntamem(uint8 *p,
	int ram,
	uint32 b)
{
	FCEUPPU_LineUpdate();

	vnapage[b] = p;
	PPUNTARAM&=~(1<<b);
	if (ram)
		PPUNTARAM|= 1<<b;
}

static int	mirrorhard = 0;

void setmirrorw(int a,
	int b,
	int c,
	int d)
{
	FCEUPPU_LineUpdate();

	vnapage[0] = NTARAM+a*0x400;
	vnapage[1] = NTARAM+b*0x400;
	vnapage[2] = NTARAM+c*0x400;
	vnapage[3] = NTARAM+d*0x400;
}

void FASTAPASS(1) setmirror(int t)
{
	FCEUPPU_LineUpdate();

	if (mirrorhard)
		return;
//	t = 0xF053>>(t<<2&12)&3^(t&12);
//	vnapage[0] = NTARAM+(t<<10&0x400);
//	vnapage[1] = NTARAM+(t<<9&0x400);
//	vnapage[2] = NTARAM+(t<<8&0x400);
//	vnapage[3] = NTARAM+(t<<7&0x400);

	switch(t)
	{
	case MI_H:
		vnapage[0] = vnapage[1] = NTARAM;
		vnapage[2] = vnapage[3] = NTARAM+0x400;
		break;
	case MI_V:
		vnapage[0] = vnapage[2] = NTARAM;
		vnapage[1] = vnapage[3] = NTARAM+0x400;
		break;
	case MI_0:
		vnapage[0] = vnapage[1] = vnapage[2] = vnapage[3] = NTARAM;
		break;
	case MI_1:
		vnapage[0] = vnapage[1] = vnapage[2] = vnapage[3] = NTARAM+0x400;
		break;
	}
	PPUNTARAM = 15;
}

void SetupCartMirroring(int m,
	int hard,
	uint8 *extra)
{
	if (m<4)
	{
		mirrorhard = 0;
		setmirror(m);
	} else
	{
		vnapage[0] = NTARAM;
		vnapage[1] = NTARAM+0x400;
		vnapage[2] = extra;
		vnapage[3] = extra+0x400;
		PPUNTARAM = 15;
	}

	mirrorhard = hard;
}

void FCEU_SaveGameSave(CartInfo *LocalHWInfo)
{
	FILE		*sp;
	char		*soot;
	int		x = 0;

	if (!LocalHWInfo->battery || LocalHWInfo->SaveGame[0]==0)
		return;
	if ((soot=FCEU_MakeFName(FCEUMKF_SAV,0,"sav"))==0)
		return;
	if ((sp=FCEUD_UTF8fopen(soot,"wb"))==0)
	{
		FCEU_PrintError("WRAM file \"%s\" cannot be written to.\n",soot);
	} else
		do
			if (LocalHWInfo->SaveGame[x])
				fwrite(LocalHWInfo->SaveGame[x],1,
					LocalHWInfo->SaveGameLen[x],sp);
		while(++x<4);
	free(soot);
}

void FCEU_LoadGameSave(CartInfo *LocalHWInfo)
{
	FILE		*sp;
	char		*soot;

	if (!LocalHWInfo->battery || LocalHWInfo->SaveGame[0]==0)
		return;

	if ((soot=FCEU_MakeFName(FCEUMKF_SAV,0,"sav"))==0)
		return;
	if ((sp=FCEUD_UTF8fopen(soot,"rb"))!=0)
	{
		int	x = 0;

		do
			if (LocalHWInfo->SaveGame[x])
				fread(LocalHWInfo->SaveGame[x],1,LocalHWInfo->SaveGameLen[x],sp);
		while(++x<4);
	}
	free(soot);
}
