/*
 macos/macfiles.c

  Created: 2008.07.17
  Revised: 2010.04.11
*/
/*
 To Do: implement the recent files as a submenu (store FSSpecs not path strings).
*/

#include <Files.h>

#include <stdio.h>
#include <stdlib.h>

#include "shell.h"

#include "../FCEU.h"

#include "shellmenus.h"
#include "macdialogs.h"
#include "macinput.h"
#include "macvideo.h"

#ifdef env68k
 #pragma segment files
#endif

/* ---- globals */

static FSSpec		gamespec;
static char		*gamepath = NULL;

extern struct FCEUGI	*GI;

/* ---- code */

#if !TARGET_API_MAC_CARBON && __MRC__
 #undef fileno
int fileno(FILE *iop);
int fileno(FILE *iop)
{
	return iop->_file;
}
#endif

#if defined(__SC__) || TARGET_API_MAC_CARBON
# include <FSpio.h>
#else
static FILE *FSp_fopen(const FSSpec *spec,
	const char *mode)
{
	Str255		name;

	FILE		*file;
	long		dir;
	short		vol;
	OSErr		err;

	if ((err=HGetVol(name,&vol,&dir))==noErr)
		HSetVol(NULL,spec->vRefNum,spec->parID);

	memcpy(name,1+spec->name,spec->name[0]);
	name[spec->name[0]] = 0;
	file = fopen((char*)name,mode);

	if (err==noErr)
		HSetVol(NULL,vol,dir);

	return file;
}
#endif

FILE *FCEUD_UTF8fopen(const char *fn,
	const char *mode)
{
#ifdef mac
	FSSpec		tmpspec;

	/* ...yeah, match the volume; 2009.09.29 */
	if (strcmp(fn,getlastfilepath(&tmpspec.vRefNum,&tmpspec.parID,tmpspec.name))==0)
		return FSp_fopen(&tmpspec,mode);
	else if (strcmp(fn,gamepath)==0)
		return FSp_fopen(&gamespec,mode);
#endif
	return fopen(fn,mode);
}

int takefile(const FSSpec *file)
{
	int		r = gamepath==NULL;

	if (gamepath!=NULL) /* clear */
		gamepath = realloc(gamepath,0);
	if ((gamepath=makepathforfsspec(file,NULL))==NULL)
		return -1;
	gamespec = *file;
	/* will open on next event */

	return r;
}

/* ---- paths */

static void addrecent(const char *lameo)
{
#pragma unused(lameo)
}

int tracknewpath(char **store,
	const char *path,
	const char *initdir,
	uint16 nameoffset,
	int priority)
{
	char		*p = NULL;

	if (path==NULL)
		return -3; //idiot proofing

	if (nameoffset!=0)
	{
		if ((p=realloc(NULL,nameoffset+1))==NULL)
			return -1;
		strncpy(p,path,nameoffset);
		p[nameoffset] = 0;
	}

	if (priority>=0)
	{
		if (priority==0)
			addrecent(path+nameoffset);
		/* add recent directory */
	}

	if (nameoffset==0)
		goto end;
	if (store!=NULL && initdir==NULL)
	{
		if (*store!=NULL)
			*store = realloc(*store,0);
		*store = p;
		return 1;
	}
	p = realloc(p,0);
end:
	return 0;
}

/* ---- game code */

const char *getnextgamepath()
{ /* const to prevent attempts to modify the string */
	return (const char*)gamepath;
}

int cleargamepath()
{
	if (scexit(0)==(_sc_reopengame|_sc_closegame))
		return 0; /* pending file... */
	if (gamepath!=NULL)
		gamepath = realloc(gamepath,0);

	return 1;
}

int loadgame(const char *nameo)
{
	scexit(~_sc_reopengame); /* make sure this's cleared; 2010.04.08 */

	if (nameo!=NULL)
	{
		void		*p;
		long		z = 1+strlen(nameo);

		if ((p=realloc(gamepath,z))==NULL)
			return -1;
		gamepath = p;
		memcpy(gamepath,nameo,z);
		nameo = gamepath;
	} else if ((nameo=gamepath)==NULL)
		return 0;

	setcursor(4,false); /* set to watch cursor; 2009.10.28 13:50 EDT */

	if ((GI=FCEUI_LoadGame(nameo))!=NULL)
	{
		addrecent(nameo);
		if (eoptions&EO_FSAFTERLOAD)
			initializescreen(true);
		videorefresh();
	} else
	{
		stopsound();
		if (nameo==gamepath && gamepath!=NULL)
			gamepath = realloc(gamepath,0);
	}

	parseinputgi(GI);
	redomenusgi(GI);
#ifdef NETWORK
	FCEUD_NetworkConnect();
#endif
	setcursor(-1,false);

	return 1;
}

void opengame(const char *initdir)
{
	static char	*gfsdir = NULL;

	const char	*filter = "All usable files(*.nes,*.nsf,*.fds,*.unf,*.zip,*.gz)\0"
"*.nes;*.nsf;*.fds;*.unf;*.zip;*.gz\0All non-compressed usable files(*.nes,*.nsf,*.fds,*.unf)\0"
"*.nes;*.nsf;*.fds;*.unf\0All files (*.*)\0*.*\0";
	uint16		nameoffset = 0;

	if (GI!=NULL)
	{ /* delayed call for reopen */
		scexit(_sc_reopengame);
		return;
	}
	if ((gamepath=getfilepath(gfp_getfile,
			"\pFCE Ultra Open File...",
			NULL,initdir==NULL ? gfsdir:initdir,NULL,
			filter,
			&nameoffset,
			gfpf_filemustexist|gfpf_hidereadonly))!=NULL)
	{
		tracknewpath(&gfsdir,gamepath,initdir,nameoffset,0);

		/* file will be automatically opened at next event */
	}
}

/* ---- state files */

static char	*state_gfsdir = NULL;

void savestate(const char *initdir)
{
	static char	currentname[256]; /* current name storage; 2009.01.25 */

	char		*filepath = NULL;
	uint16		nameoffs = 0, length = 0;

	if (currentname[0]==0)
		getcstr(currentname,strlistid_common,commonstr_defaultstatename);

	if ((filepath=getfilepath(gfp_putfile,
			"\pSave State As...",
			NULL,initdir==NULL ? state_gfsdir:initdir,currentname,
			"FCE Ultra Save State(*.fc?)\0*.fc?\0",
			&nameoffs,
			gfpf_filemustexist|gfpf_hidereadonly))!=NULL)
	{
		FCEUI_SaveState(filepath);

		tracknewpath(&state_gfsdir,filepath,initdir,nameoffs,7);

		if (nameoffs>0 && (unsigned)(length=1+strlen(filepath)-nameoffs)<=256)
			memcpy(currentname,nameoffs+filepath,length);
	}

	if (filepath!=NULL)
		filepath = realloc(filepath,0);
}

void loadstate(const char *initdir)
{
	char		*filepath = NULL;
	uint16		nameoffs = 0;

	if ((filepath=getfilepath(gfp_getfile,
			"\pLoad State From...",
			NULL,initdir==NULL ? state_gfsdir:initdir,NULL,
			"FCE Ultra Save State(*.fc?)\0*.fc?\0",
			&nameoffs,
			gfpf_filemustexist|gfpf_hidereadonly))!=NULL)
	{
		FCEUI_LoadState(filepath);

		tracknewpath(&state_gfsdir,filepath,initdir,nameoffs,1);
	}

	if (filepath!=NULL)
		filepath = realloc(filepath,0);
}

/* ---- sound log */

int recordsoundwave(const char *initdir)
{
	static char	*gfsdir = NULL;

	static char	currentname[256]; /* current name storage; 2009.01.25 */

	char		*filepath = NULL;
	int		result = 0;
	uint16		nameoffs = 0, length = 0;

	if (currentname[0]==0)
		getcstr(currentname,strlistid_common,commonstr_defaultwavename);

	if (FCEUI_EndWaveRecord())
		return 0; /* select again to start again (no option to stop without dialog?) */

	if ((filepath=getfilepath(gfp_putfile,
			"\pLog Sound As...",
			NULL,initdir==NULL ? gfsdir:initdir,currentname,
			"MS WAVE(*.wav)\0*.wav\0",
			&nameoffs,
			gfpf_filemustexist|gfpf_hidereadonly))!=NULL)
	{
		result = FCEUI_BeginWaveRecord(filepath);

		tracknewpath(&gfsdir,filepath,initdir,nameoffs,9);

		if (nameoffs>0 && (unsigned)(length=1+strlen(filepath)-nameoffs)<=256)
			memcpy(currentname,nameoffs+filepath,length);
	}

	if (filepath!=NULL)
		filepath = realloc(filepath,0);

	return result;
}
