/*
 movie.c

 * MacFCEU 0.6a4
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "driver.h"
#include "FCEU.h"
#include "input.h"
#include "movie.h"
#include "state.h"
#include "general.h"
#include "video.h"

#ifdef env68k
 #pragma segment files
#endif

static int	current = 0;		// > 0 for recording, < 0 for playback
static FILE	*slots[10] = {0};
static uint8	joop[4];
static uint32	framets;

/* Cache variables used for playback. */
static uint32	nextts;
static int	nextd;


static int	CurrentMovie = 0;
static int	MovieShow = 0;

static int	MovieStatus[10];

static void DoEncode(int joy, int button, int dummy)
{
	uint8		d;

	d = 0;

	if (framets>=65536)
		d = 3<<5;
	else if (framets>=256)
		d = 2<<5;
	else if (framets>0)
		d = 1<<5;

	if (dummy!=0)
		d|= 0x80;

	d|= joy<<3;
	d|= button;

	fputc(d,slots[current-1]);
	//printf("Wr: %02x, %d\n",d,slots[current-1]);
	if (framets!=0)
		do
			fputc(framets&0xFF,slots[current-1]);
		while((framets>>=8)!=0);
}

int FCEUMOV_IsPlaying(void)
{
	return current<0 ? 1:0;
}

void FCEUI_SaveMovie(const char *fname)
{
	FILE		*fp;
	char		*fn;

	if (current<0)	/* Can't interrupt playback.*/
		return;

	if (current>0)	/* Stop saving. */
	{
		/* Write a dummy timestamp value so that the movie will keep
			"playing" after user input has stopped.
		*/
		DoEncode(0,0,1);
		fclose(slots[--current]);
		MovieStatus[current] = 1;
		current = 0;
		FCEU_DispMessage("Movie recording stopped.");
		return;
	}

	current = CurrentMovie;

	if (fname)
		fp = FCEUD_UTF8fopen(fname, "wb");
	else if ((fn=FCEU_MakeFName(FCEUMKF_MOVIE,CurrentMovie,0))!=0)
	{
		fp = FCEUD_UTF8fopen(fn,"wb");
		free(fn);
	} else
		return;

	if (fp==0)
		return;

	FCEUSS_SaveFP(fp);
	fseek(fp, 0, SEEK_END);
	slots[current++] = fp;
	memset(joop,0,sizeof(joop));

	framets = 0;
	FCEUI_SelectMovie(CurrentMovie); //Quick hack to display status.
}

static void StopPlayback(void)
{
	fclose(slots[~current]); /*-1 - */
	current = 0;
	FCEU_DispMessage("Movie playback stopped.");
}

void FCEUMOV_Stop(void)
{
	if (current<0)
		StopPlayback();
}

void FCEUI_LoadMovie(const char *fname)
{
	FILE		*fp;
	char		*fn;

	if (current>0)	/* Can't interrupt recording.*/
		return;

	if (current<0)	/* Stop playback. */
	{
		StopPlayback();
		return;
	}

	if (fname)
		fp = FCEUD_UTF8fopen(fname, "rb");
	else if ((fn=FCEU_MakeFName(FCEUMKF_MOVIE,CurrentMovie,0))!=0)
	{
		fp = FCEUD_UTF8fopen(fn,"rb");
		free(fn);
	} else
		return;

	if (fp==0)
		return;

	if (!FCEUSS_LoadFP(fp))
		return;

	current = CurrentMovie;
	slots[current] = fp;

	memset(joop,0,sizeof(joop));
	current = ~current; /*-1 - */
	framets = 0;
	nextts = 0;
	nextd = -1;
	MovieStatus[CurrentMovie] = 1;
	FCEUI_SelectMovie(CurrentMovie); /* Quick hack to display status. */
}

void FCEUMOV_AddJoy(uint8 *js)
{
	int		x, y;

	if (!current)
		return;	/* Not playback nor recording. */

	if (current<0)	/* Playback */
	{
		while(nextts==framets)
		{
			int		tmp, ti;
			uint8		d;

			if (nextd!=-1)
				if (nextd&0x80)
				{
					//puts("Egads");
					FCEU_DoSimpleCommand(nextd&0x1F);
				} else
					joop[nextd>>3&3]^= 1<<(nextd&7);


			/*-1 - */
			if ((tmp=fgetc(slots[~current]))<0)
			{
				StopPlayback();
				return;
			}
			d = tmp;

			nextts = 0;
			tmp >>= 5;
			tmp &= 0x3;
			ti = 0;

			{
				int	tmpfix = tmp;

				while(tmp--!=0)
				{ /*-1 - */
					nextts|= fgetc(slots[~current])<<(ti * 8);
					++ti;
				}

	// This fixes a bug in movies recorded before version 0.98.11
	// It's probably not necessary, but it may keep away CRAZY PEOPLE who recorded
	// movies on <= 0.98.10 and don't work on playback.
				if (tmpfix==1 && !nextts)
					nextts|= fgetc(slots[~current])<<8;
				else if (tmpfix==2 && !nextts)
					nextts|= fgetc(slots[~current])<<16;

				framets = 0;
				nextd = d;
			}

		}
		memcpy(js,joop,4);
	} else	/* Recording */
		for(x=0; x<4; ++x)
		{
			if (js[x]!=joop[x])
			{
				for(y=0;y<8;y++)
					if ((js[x]^joop[x])&(1<<y))
						DoEncode(x,y,0);
				joop[x] = js[x];
			} else if (framets==((1<<24)-1))
			{ /* Overflow will happen, so do dummy update. */
				DoEncode(0,0,1);
			}
		}

	++framets;
}

void FCEUMOV_AddCommand(int cmd)
{
	if (current<=0)
		return;	/* Return if not recording a movie */
	//printf("%d\n",cmd);
	DoEncode((cmd>>3)&3,cmd&7,1);
}

void FCEUMOV_CheckMovies(void)
{
	FILE		*st = 0;
	char		*fn;
	int		ssel;

	for(ssel=0; ssel<10; ++ssel)
	{
		if ((fn=FCEU_MakeFName(FCEUMKF_MOVIE,ssel,0))==0)
			continue;
		st = FCEUD_UTF8fopen(fn,"rb");
		free(fn);
		if (st)
		{
			MovieStatus[ssel] = 1;
			fclose(st);
		} else
			MovieStatus[ssel] = 0;
	}
}

void FCEUI_SelectMovie(int w)
{
	if (w==-1)
		{MovieShow = 0; return;}
	FCEUI_SelectState(-1);

	CurrentMovie = w;
	MovieShow = 180;

	if (current>0)
		FCEU_DispMessage("-recording movie %d-",current-1);
	else if (current<0)
		FCEU_DispMessage("-playing movie %d-",~current); /*-1 - */
	else
		FCEU_DispMessage("-select movie-");
}

void FCEU_DrawMovies(uint8 *XBuf)
{
	if (MovieShow==0)
		return;

	FCEU_DrawNumberRow(XBuf,MovieStatus, CurrentMovie);
	--MovieShow;
}
