/*
 sound.c
  base for N2A0x pseuedo-Audio Processing Unit emulation

 * MacFCEU 0.6.1
 *
 * FCE Ultra - NES/Famicom Emulator
 *
 * Copyright notice for this file:
 *  sound.c Copyright (C) 2002 Xodnizel
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <string.h>

#include "driver.h"
#include "x6502.h"
#include "FCEU.h"
#include "WAVE.h"
#include "fsound.h"
#include "filter.h"
#include "state.h"

#include "utility/ftrig.h"

#ifdef env68k
 #pragma segment N2A0x
#endif

/* amplitude tables */
static int32	rectwlookup[32];
static int32	tpcmwlookup[203]; //triangle plain/noise, PCM

/* virtual wave, prior to filtering */
int32		FCEUWave[12000]; /*2048+512 -- increased for higher sample rates, lower speeds */
int32		FCEUWaveHi[40000];

/* post-filter results */
static int16	WaveFinal[12000]; /*2048+512*/

EXPSOUND	GameExpSound = {0,0,0};

static uint8	TriCount;
static uint8	TriMode;

static int32	tristep;

static int32	wlcount[4];		/* Wave length counters. */

static uint8	PSG[0x10];

uint8		IRQFrameMode;		/* $4017 / xx000000 */
uint8		RawDALatch;		/* $4011 0xxxxxxx */

uint8		EnabledChannels;	/* Byte written to $4015 */

typedef struct ENVUNIT/*(8)*/
{
	uint8	Speed;
	uint8	Mode;	/* Fixed volume(1), and loop(2) */
	uint8	DecCountTo1;
	uint8	decvolume;
	int	reloaddec;
} ENVUNIT;

static ENVUNIT	EnvUnits[3];

//static const int32 RectDuties[4] = {1,2,4,6};

static int32	RectDutyCount[2];
static uint8	sweepon[2];
static int32	curfreq[2];
static uint8	SweepCount[2];

static uint16	nreg;

uint8		Soundfcnt;
int32		Soundfhcnt;
int32		Soundfhinc;

uint32		soundtsoffs;

/* Variables exclusively for low-quality sound. */
int32		nesincsize;
uint32		soundtsinc;
uint32		soundtsi;
static int32	sqacc[2];
/* LQ variables segment ends. */

int32		soundlencount[4];

static const uint8 lengthtable[0x20] =
{ /* updated to fit apu_ref.txt - 0x0D*2 (0x1A) on second row was 0x0e*2 */
	0x0A, 0xFE, 0x14, 0x02, 0x28, 0x04, 0x50, 0x06,
	0xA0, 0x08, 0x3C, 0x0A, 0x0E, 0x0C, 0x1A, 0x0E,
	0x0C, 0x10, 0x18, 0x12, 0x30, 0x14, 0x60, 0x16,
	0xC0, 0x18, 0x48, 0x1A, 0x10, 0x1C, 0x20, 0x1E
};

static const uint32 NTSCNoiseFreqTable[0x10] =
{
	  4,   8,  16,  32,  64,  96, 128, 160,
	202, 254, 380, 508, 762,1016,2034,4068
};

static const uint32 PALNoiseFreqTable[0x10] =
{
	  4,   7,  14,  30,  60,  88, 118, 148,
	188, 236, 354, 472, 708, 944,1890,3778
};

static const uint32 NTSCDMCTable[0x10] =
{
	428, 380, 340, 320, 286, 254, 226, 214, /* 340 should be 336? */
	190, 160, 142, 128, 106,  84,  72,  54
};

/* exactly 1/1.08 that of the NTSC table (rounded twice) */
static const uint32 PALDMCTable[0x10] =
{
	398, 354, 316, 298, 266, 236, 210, 198,
	176, 148, 131, 118,  98,  78,  66,  50,
};

/* one of the two tables above loaded by SetSoundVariables */
static uint32 CurrentNoiseFreqTable[0x10];
static uint32 CurrentDMCTable[0x10];

// $4010        -        Frequency
// $4011        -        Actual data outputted
// $4012        -        Address register: $c000 + V*64
// $4013        -        Size register:  Size in bytes = (V+1)*64

int32		DMCacc = 1;
int32		DMCPeriod;
uint8		DMCBitCount = 0;

uint8		DMCAddressLatch, DMCSizeLatch; /* writes to 4012 and 4013 */
uint8		DMCFormat;	/* Write to $4010 */

uint32		DMCAddress = 0;
int32		DMCSize = 0;
uint8		DMCShiftF[2] = {0}; /* first byte kept zero for BE 68K Asm */
#define DMCShift DMCShiftF[1]

uint8		SIRQStat = 0;

uint8		DMCDMABuf = 0;
uint8		DMCHaveDMA = 0;
uint8		DMCHaveSample = 0;

static void Dummyfunc(void) {};
void (*DoSQ1)(void)=Dummyfunc;
void (*DoSQ2)(void)=Dummyfunc;
void (*DoTriangle)(void)=Dummyfunc;
void (*DoNoise)(void)=Dummyfunc;
void (*DoPCM)(void)=Dummyfunc;

static uint32	ChannelBC[5];

/*static void LoadDMCPeriod(uint8 V)*/
#define LoadDMCPeriod(V) \
	(void)(DMCPeriod=PAL ? PALDMCTable[(uint8)(V)]:CurrentDMCTable[(uint8)(V)])

#define PrepDPCM() \
	do{DMCAddress=0x4000+(DMCAddressLatch<<6),DMCSize=(DMCSizeLatch<<4)+1;}while(0)

/* Instantaneous?  Maybe the new freq value is being calculated all of the time... */

static int FASTAPASS(2) CheckFreq(uint32 cf, uint8 sr)
{
	return ((sr&8)==0 && ((cf>>(sr&7))+cf)&0x800) ? 0:1;
}

static void SQReload(int x, uint8 V)
{
	if (EnabledChannels&(x+1)) /*(1<<x)*/
	{
		if (x!=0)
			DoSQ2();
		else
			DoSQ1();
		soundlencount[x] = lengthtable[V>>3&31];
	}

	sweepon[x] = PSG[x*4+1]&0x80;
	curfreq[x] = PSG[x*4+2]|(V<<8&0x700);
	SweepCount[x] = (PSG[x*4+1]>>4&7)+1;

	RectDutyCount[x] = 7;
	EnvUnits[x].reloaddec = 1;
}

static DECLFW(Write_PSG)
{
	A&= 31;
	switch(A)
	{
	case 0x0:
		DoSQ1();
		EnvUnits[0].Mode = V>>4&3;
		EnvUnits[0].Speed = V&15;
		break;
	case 0x1:
		sweepon[0] = V&0x80;
		break;
	case 0x2:
		DoSQ1();
		curfreq[0]&= 0xFF00;
		curfreq[0]|= V;
		break;
	case 0x3:
		SQReload(0,V);
		break;
	case 0x4:
		DoSQ2();
		EnvUnits[1].Mode = V>>4&3;
		EnvUnits[1].Speed = V&15;
		break;
	case 0x5:
		sweepon[1] = V&0x80;
		break;
	case 0x6:
		DoSQ2();
		curfreq[1]&= 0xFF00;
		curfreq[1]|= V;
		break;
	case 0x7:
		SQReload(1,V);
		break;
	case 0xA:
		DoTriangle();
		break;
	case 0xB:
		DoTriangle();
		if (EnabledChannels&4)
			soundlencount[2] = lengthtable[V>>3&31];
		TriMode = 1; /* Load mode */
		break;
	case 0xC:
		DoNoise();
		EnvUnits[2].Mode = V>>4&3;
		EnvUnits[2].Speed = V&15;
		break;
	case 0xE:
		DoNoise();
		break;
	case 0xF:
		DoNoise();
		if (EnabledChannels&8)
			soundlencount[3] = lengthtable[V>>3&31];
		EnvUnits[2].reloaddec = 1;
		break;
	case 0x10:
		DoPCM();
		LoadDMCPeriod(V&15);

		if (SIRQStat&0x80)
			if ((V&0x80)==0)
			{
				XFrame.IRQlow&=~FCEU_IQDPCM; /*X6502_IRQEnd(FCEU_IQDPCM);*/
				SIRQStat&=~0x80;
			} else
				XFrame.IRQlow|= FCEU_IQDPCM; /*X6502_IRQBegin(FCEU_IQDPCM);*/
		break;
	}
	PSG[A] = V;
}

static DECLFW(Write_DMCRegs)
{
	if ((A&=15)==0) //4010h
	{
		DoPCM();
		LoadDMCPeriod(V&15);

		if (SIRQStat&0x80)
		{
			if ((V&0x80)==0)
			{
				XFrame.IRQlow&=~FCEU_IQDPCM; //X6502_IRQEnd
				SIRQStat&=~0x80;
			} else
				XFrame.IRQlow|= FCEU_IQDPCM; //X6502_IRQBegin
		}
		DMCFormat = V;
	} else if (--A==0) //4011h
	{
		DoPCM();
		RawDALatch = V&0x7F;
	} else if (--A==0) //4012h
		DMCAddressLatch = V;
	else if (--A==0) //4013h
		DMCSizeLatch = V;
}

#ifdef env68k
DECLFW(SndStatusWrite);
DECLFR(SndStatusRead);
#else
static DECLFW(SndStatusWrite) //4015h
{
(void)A;
	DoSQ1();
	DoSQ2();
	DoTriangle();
	DoNoise();
	DoPCM();

	/* Force length counters to 0. */
	if ((V&1)==0)
		soundlencount[0] = 0;
	if ((V&2)==0)
		soundlencount[1] = 0;
	if ((V&4)==0)
		soundlencount[2] = 0;
	if ((V&8)==0)
		soundlencount[3] = 0;

	if ((V&0x10)==0)
		DMCSize = 0;
	else if (DMCSize==0)
		PrepDPCM();

	SIRQStat&=~0x80;
	XFrame.IRQlow&=~FCEU_IQDPCM; /*X6502_IRQEnd(FCEU_IQDPCM);*/
	EnabledChannels = V&31;
}

static DECLFR(SndStatusRead)
{
	uint8		ret;
(void)A;
	ret = DMCSize!=0, ret+= ret;
	ret|= soundlencount[3]!=0, ret+= ret;
	ret|= soundlencount[2]!=0, ret+= ret;
	ret|= soundlencount[1]!=0, ret+= ret;
	ret|= soundlencount[0]!=0;
	ret|= SIRQStat;

	if (!fceuindbg)
	{
		SIRQStat&=~0x40;
		XFrame.IRQlow&=~FCEU_IQFCOUNT; /*X6502_IRQEnd(FCEU_IQFCOUNT);*/
	}

	return ret;
}
#endif

#define sweepIx(P) do{ \
	if((PSG[P<<2]&0x20)==0) /* Make sure loop flag is not set. */ \
		if(soundlencount[P]>0) \
			--soundlencount[P]; \
\
	/* Frequency Sweep Code Here */ \
	/* xxxx 0000 */ \
	/* xxxx = hz.  120/(P+1) */ \
	if(sweepon[P]!=0) {int32 mod=0; \
		if(SweepCount[P]>0) \
			--SweepCount[P];  \
		if(SweepCount[P]<=0) { \
			SweepCount[P]=(PSG[(P<<2)+1]>>4&7)+1; /*+1;*/ \
		if(PSG[(P<<2)+0x1]&8) { \
			mod-=(P^1)+(curfreq[P]>>(PSG[(P<<2)+1]&7)); \
			if(curfreq[P]!=0 && (PSG[(P<<2)+1]&7) /*&& sweepon[P]&0x80*/ ) \
				curfreq[P]+=mod; \
			} else { \
				mod=curfreq[P]>>(PSG[(P<<2)+1]&7); \
				if((mod+curfreq[P])&0x800) \
					sweepon[P]=0,curfreq[P]=0; \
				else if(curfreq[P] && (PSG[(P<<2)+1]&7) /*&& sweepon[P]&0x80*/ ) \
					curfreq[P]+=mod; \
			} \
		} \
	} else { /* Sweeping is disabled: */ \
		/*curfreq[P]&=0xFF00; curfreq[P]|=PSG[(P<<2)|2]; //|((PSG[(P<<2)|3]&7)<<8);*/ \
	} \
}while(0)

#define envelopeDecayIx(P) do{ \
	if(EnvUnits[P].reloaddec) { \
		EnvUnits[P].decvolume=15; EnvUnits[P].DecCountTo1=EnvUnits[P].Speed+1; \
		EnvUnits[P].reloaddec=0; \
		continue; \
	} \
	if(EnvUnits[P].DecCountTo1>0) --EnvUnits[P].DecCountTo1; \
	if(EnvUnits[P].DecCountTo1==0) { \
		EnvUnits[P].DecCountTo1=EnvUnits[P].Speed+1; \
		if(EnvUnits[P].decvolume || (EnvUnits[P].Mode&2)) \
			EnvUnits[P].decvolume--,EnvUnits[P].decvolume&=15; \
	} \
}while(0)

#ifdef env68k
void FASTAPASS(1) FrameSoundStuff(int V);
#else
static
#endif
void FASTAPASS(1) FrameSoundStuff(int V)
{
	DoSQ1();
	DoSQ2();
	DoNoise();
	DoTriangle();

	if ((V&1)==0) /* Envelope decay, linear counter, length counter, freq sweep */
	{
		if ((PSG[8]&0x80)==0)
			if (soundlencount[2]>0)
				--soundlencount[2];
		if ((PSG[0xC]&0x20)==0)	/* Make sure loop flag is not set. */
			if (soundlencount[3]>0)
				--soundlencount[3];

		sweepIx(0);
		sweepIx(1);
		sweepIx(2);
	}

	/* Now do envelope decay + linear counter. */

	if (TriMode!=0) // In load mode?
		TriCount = PSG[8]&0x7F;
	else if (TriCount!=0)
		--TriCount;

	if ((PSG[8]&0x80)==0)
		TriMode = 0;

	envelopeDecayIx(0);
	envelopeDecayIx(1);
	envelopeDecayIx(2);
}

#undef sweepIx
#undef envelopeDecayIx

#ifndef env68k /* a number of functions written into 2A0x.a */
void FrameSoundUpdate(void)
{
 // Linear counter:  Bit 0-6 of $4008
 // Length counter:  Bit 4-7 of $4003, $4007, $400b, $400f

	if (Soundfcnt==0 && (IRQFrameMode&3)==0)
	{
		SIRQStat|= 0x40;
		XFrame.IRQlow|= FCEU_IQFCOUNT; /*X6502_IRQBegin*/
	}

	if (Soundfcnt==3)
		if (IRQFrameMode&2)
			Soundfhcnt+= Soundfhinc;
	FrameSoundStuff(Soundfcnt);
	Soundfcnt = (Soundfcnt+1)&3;
}

static INLINE void DMCDMA(void)
{
	if (DMCSize && !DMCHaveDMA)
	{
		X6502_DMR(0x8000+DMCAddress);
		X6502_DMR(0x8000+DMCAddress);
		X6502_DMR(0x8000+DMCAddress);
		DMCDMABuf = X6502_DMR(0x8000+DMCAddress);
		DMCHaveDMA = 1;
		DMCAddress = (DMCAddress+1)&0x7fff;

		if (--DMCSize==0)
			if (DMCFormat&0x40)
				PrepDPCM();
			else
			{
				SIRQStat|= 0x80;
				if (DMCFormat&0x80)
					XFrame.IRQlow|= FCEU_IQDPCM; /*X6502_IRQBegin*/
			}
	}
}

void FASTAPASS(1) FCEU_SoundCPUHook(int cycles)
{
	if ((Soundfhcnt-=cycles*48)<=0)
	{
		FrameSoundUpdate();
		Soundfhcnt+= Soundfhinc;
	}

	DMCDMA();
	DMCacc-= cycles;

	while(DMCacc<=0)
	{
		unsigned	shift = DMCShift;

		DMCShift = shift>>1;
		if (DMCHaveSample)
		{
			uint8	bah = RawDALatch;

			if (FSettings.SndRate!=0)
			{ /* Unbelievably ugly hack */
				soundtsoffs+= DMCacc;
				DoPCM();
				soundtsoffs-= DMCacc;
			}

			if ((RawDALatch+=(shift<<2&4)-2)&0x80)
				RawDALatch = bah;
		}

		DMCacc+= DMCPeriod;
		if ((DMCBitCount=(DMCBitCount+1)&7)==0)
			if ((DMCHaveSample=DMCHaveDMA)!=0)
				DMCShift = DMCDMABuf, DMCHaveDMA = 0;
	}
}
#endif

void RDoPCM(void);
void RDoPCM(void)
{
	uint32		V;

	for(V=ChannelBC[4]; V<SOUNDTS; ++V)
		FCEUWaveHi[V]+= RawDALatch<<16;

	ChannelBC[4] = SOUNDTS;
}

/* This has the correct phase.  Don't mess with it. */
static INLINE void RDoSQ(int x)
{
	int32		V;
	int32		amp;
	int32		rthresh;
	int32		*D;
	int32		currdc;
	int32		cf;
	int32		rc;

	if (curfreq[x]<8 || curfreq[x]>0x7FF)
		goto endit;
	if (CheckFreq(curfreq[x],PSG[x*4+1])==0)
		goto endit;
	if (soundlencount[x]==0)
		goto endit;

	if (EnvUnits[x].Mode&1)
		amp = EnvUnits[x].Speed;
	else
		amp = EnvUnits[x].decvolume;
	amp<<= 24;

	//rthresh = RectDuties[PSG[x*4]>>6];
	rthresh = PSG[x*4]>>6;
	if ((rthresh+=rthresh)==0)
		rthresh = 1;

	D = &FCEUWaveHi[ChannelBC[x]];
	V = SOUNDTS-ChannelBC[x];

	currdc = RectDutyCount[x];
	cf = (curfreq[x]+1)*2;
	rc = wlcount[x];

	if (V>0)
		do
		{
			if (currdc<rthresh)
				*D+= amp;
			if (--rc==0)
				rc = cf, currdc = (currdc+1)&7;
		} while(++D,--V!=0);

	RectDutyCount[x] = currdc;
	wlcount[x] = rc;
endit:
	ChannelBC[x] = SOUNDTS;
}

static void RDoSQ1(void)
{
	RDoSQ(0);
}

static void RDoSQ2(void)
{
	RDoSQ(1);
}

#define makettable(x) do{int y; \
	inie[x]=nesincsize; \
	if(curfreq[x]<8 || curfreq[x]>0x7FF) \
		inie[x]=0; \
	if(CheckFreq(curfreq[x],PSG[(x<<2)|1])==0) \
		inie[x]=0; \
	if(soundlencount[x]==0 || (FSettings.soundactivech&(x+1))==0) \
		inie[x]=0; \
	if(EnvUnits[x].Mode&1) \
		amp[x]=EnvUnits[x].Speed; \
	else \
		amp[x]=EnvUnits[x].decvolume; \
	if(!inie[x]) amp[x]=0;	/* Correct? Buzzing in MM2, others otherwise... */ \
	/*rthresh[x]=RectDuties[PSG[x*4]>>6];*/ \
	y=PSG[x*4]>>6; \
	if((y+=y)==0) y=1; \
	rthresh[x]=y; \
	y=0; \
	do{ \
		if(y<rthresh[x]) \
			ttable[x][y]=amp[x]; \
		else \
			ttable[x][y]=0; \
	} while(++y<8); \
	freq[x]=curfreq[x]+1<<1+17; \
} while(0)

static void RDoSQLQ(void)
{
	int32		start, end;
	int32		V;
	int32		amp[2];
	int32		rthresh[2];
	int32		freq[2];
	int32		inie[2];

	int32		ttable[2][8];
	int32		totalout;

	start = ChannelBC[0];
	if ((end=(SOUNDTS<<16)/soundtsinc)<=start)
		return;
	ChannelBC[0] = end;

	/* unrolled */
	makettable(0);
	makettable(1);

	totalout = rectwlookup[ ttable[0][RectDutyCount[0]] + ttable[1][RectDutyCount[1]] ];
	V = start;
	if ((inie[0]|inie[1])==0)
	{
		do{FCEUWave[V>>4]+=totalout;} while(++V<end);
		RectDutyCount[0]&= 7, RectDutyCount[1]&= 7;
		return;
	}

	do
	{
		//int tmpamp=0;
		//if(RectDutyCount[0]<rthresh[0])
		// tmpamp=amp[0];
		//if(RectDutyCount[1]<rthresh[1])
		// tmpamp+=amp[1];
		//tmpamp=wlookup1[tmpamp];
		//tmpamp = wlookup1[ ttable[0][RectDutyCount[0]] + ttable[1][RectDutyCount[1]] ];

		FCEUWave[V>>4]+= totalout; //tmpamp;

		sqacc[1]-= inie[1];
		if ((sqacc[0]-=inie[0])<=0)
		{
			do
				RectDutyCount[0] = (RectDutyCount[0]+1)&7;
			while((sqacc[0]+=freq[0])<=0);

			if (sqacc[1]<=0)
				goto second; /* dodge unnecessary lookup; 2009.12.27 09:43 EST */
			totalout = rectwlookup[ ttable[0][RectDutyCount[0]]
				+ttable[1][RectDutyCount[1]] ];
			continue;
		}

		if (sqacc[1]<=0)
		{
		second:
			do
				RectDutyCount[1] = (RectDutyCount[1]+1)&7;
			while((sqacc[1]+=freq[1])<=0);

			totalout = rectwlookup[ ttable[0][RectDutyCount[0]]
				+ttable[1][RectDutyCount[1]] ];
		}
	} while(++V<end);
}

static void RDoTriangle(void)
{
	uint32		V;
	int32		tcout;

	tcout = tristep&15;
	if ((tristep&0x10)==0)
		tcout^= 15;
	tcout = (tcout*3)<<16;	//(tcout<<1);

	if (soundlencount[2]==0 || TriCount==0)
	{ /* Counter is halted, but we still need to output. */
		int32	*start = &FCEUWaveHi[ChannelBC[2]];
		int32	count;

		if ((count=SOUNDTS-ChannelBC[2])!=0)
			do
				*start+++= tcout;
			while(--count!=0);
		goto end;
	}

	for(V=ChannelBC[2]; V<SOUNDTS; ++V)
	{
		FCEUWaveHi[V]+= tcout;
		if (--wlcount[2]==0)
		{
			wlcount[2] = (PSG[0xA]|((PSG[0xB]&7)<<8))+1;
			tcout = ++tristep&15;
			if ((tristep&0x10)==0)
				tcout^= 15;
			tcout = (tcout*3)<<16;
		}
	}
end:
	ChannelBC[2] = SOUNDTS;
}

/* noise channel's always the slowest (optimize as much as possible) */
static void RDoTriangleNoisePCMLQ(void)
{
	static uint32	tcout = 0;
	static int32	triacc = 0;
	static int32	noiseacc = 0;

	int32		V;
	int32		start, end;
	int32		freq[2];
	int32		inie[2];
	uint32		amptab[2];
	uint32		noiseout, dmcout;
	int		nshift;

	int32		totalout;

	start = ChannelBC[2];
	if ((end=(SOUNDTS<<16)/soundtsinc)<=start)
		return;
	ChannelBC[2] = end;

	inie[0] = inie[1] = nesincsize;

	if ((freq[0]=(PSG[0xA]|(PSG[0xB]<<8&0x700))+1)<=4
			|| soundlencount[2]==0 || TriCount==0
			|| (FSettings.soundactivech&4)==0)
		inie[0] = 0;

	freq[0]<<= 16+1;

	if (EnvUnits[2].Mode&1)
		amptab[0] = 2*EnvUnits[2].Speed;
	else
		amptab[0] = 2*EnvUnits[2].decvolume;
	amptab[1] = 0;

	if (soundlencount[3]==0 || (FSettings.soundactivech&8)==0)
		amptab[0] = inie[1] = 0; /* Quick hack speedup, set inie[1] to 0 */

	noiseout = amptab[nreg>>14&1];

	if (PSG[0xE]&0x80)
		nshift = 8;
	else
		nshift = 13;

	if ((FSettings.soundactivech&0x10)==0)
		dmcout = 0;
	else
		dmcout = RawDALatch;
	totalout = tpcmwlookup[tcout+noiseout+dmcout];

	if (inie[0]!=0 && inie[1]!=0)
	{
		V = start;
		do
		{
			FCEUWave[V>>4]+= totalout;

			triacc-= inie[0];
			noiseacc-= inie[1];

			if (triacc<=0)
			{
				do
					++tristep, tristep&= 31;
				while((triacc+=freq[0])<=0);

				tcout = tristep&15;
				if ((tristep&0x10)==0)
					tcout^= 15;
				tcout*= 3;
				totalout = tpcmwlookup[tcout+noiseout+dmcout];
			}
			if (noiseacc<=0)
			{
				do
				{
					nreg = nreg*2+(((nreg>>nshift)^(nreg>>14))&1);
					nreg&= 0x7fff;
					noiseout = amptab[nreg>>14];
				} while((noiseacc+=CurrentNoiseFreqTable[PSG[0xE]&15]<<16+2)<=0);
				totalout = tpcmwlookup[tcout+noiseout+dmcout];
			}
		} while(++V<end);
	} else if (inie[0]!=0)
	{
		V = start;
		do
		{
			FCEUWave[V>>4]+= totalout;
			if ((triacc-=inie[0])<=0)
			{
				do
					++tristep, tristep&= 31;
				while((triacc+=freq[0])<=0);

				tcout = tristep&15;
				if ((tristep&0x10)==0)
					tcout^= 15;
				tcout*= 3;
				totalout = tpcmwlookup[tcout+noiseout+dmcout];
			}
		} while(++V<end);
	} else if (inie[1]!=0)
	{
		V = start;
		do
		{
			FCEUWave[V>>4]+= totalout;
			noiseacc-= inie[1];
			if (noiseacc<=0)
			{
				do
				{
					nreg = nreg*2+(((nreg>>nshift)^(nreg>>14))&1);
					nreg&= 0x7FFF;
					noiseout = amptab[(nreg>>14)];
				} while((noiseacc+=CurrentNoiseFreqTable[PSG[0xE]&15]<<16+2)<=0);
				totalout = tpcmwlookup[tcout+noiseout+dmcout];
			}
		} while(++V<end);
	} else
	{
		V = start;
		do
			FCEUWave[V>>4]+= totalout;
		while(++V<end);
	}
}

static void RDoNoise(void)
{
	int32		V;
	int32		outo;
	uint32		amptab[2];

	if (EnvUnits[2].Mode&1)
		amptab[0] = EnvUnits[2].Speed;
	else
		amptab[0] = EnvUnits[2].decvolume;

	amptab[0]<<= 16+1;
	amptab[1] = 0;

	outo = amptab[nreg&1]; //(nreg>>0xe)&1];

	if (!soundlencount[3])
		outo = amptab[0] = 0;

	if (PSG[0xE]&0x80) /* "short" noise */
		for(V=ChannelBC[3]; V<SOUNDTS; ++V)
		{
			FCEUWaveHi[V]+= outo;
			if (--wlcount[3]==0)
			{
				uint8	feedback;

				wlcount[3] = CurrentNoiseFreqTable[PSG[0xE]&15]*2;
				feedback = (nreg>>8&1)^(nreg>>14&1); /* short */
				nreg+= nreg, nreg+= feedback;
				nreg&= 0x7FFF;
				outo = amptab[nreg>>14&1];
			}
		}
	else
		for(V=ChannelBC[3]; V<SOUNDTS; ++V)
		{
			FCEUWaveHi[V]+= outo;
			if (--wlcount[3]==0)
			{
				uint8	feedback;

				wlcount[3] = CurrentNoiseFreqTable[PSG[0xE]&15]*2;
				feedback = (nreg>>13&1)^(nreg>>14&1); /* long */
				nreg+= nreg, nreg+= feedback;
				nreg&= 0x7FFF;
				outo = amptab[nreg>>14&1];
			}
		}
	ChannelBC[3] = SOUNDTS;
}

static DECLFW(Write_IRQFM) //4017h
{
(void)A;
	Soundfcnt = 0;
	if ((V>>=6)&2)
		FrameSoundUpdate();
	Soundfcnt = 1;
	Soundfhcnt = Soundfhinc;

	XFrame.IRQlow&=~FCEU_IQFCOUNT; /*X6502_IRQEnd(FCEU_IQFCOUNT);*/
	SIRQStat&= ~0x40;
	IRQFrameMode = V;
}

void SetNESSoundMap(void)
{
	SetWriteHandler(0x4000,0x400F,Write_PSG);
	SetWriteHandler(0x4010,0x4013,Write_DMCRegs);
	SetWriteHandler(0x4017,0x4017,Write_IRQFM);

	SetWriteHandler(0x4015,0x4015,SndStatusWrite);
	SetReadHandler(0x4015,0x4015,SndStatusRead);
}

static int32	inbuf = 0;

int FlushEmulateSound(void)
{
	int		x;
	int32		end, left;

	if (timestamp==0)
		return 0;

	if (FSettings.SndRate==0)
	{
		left = 0, end = 0;
		goto nosoundo;
	}

	DoSQ1();
	DoSQ2();
	DoTriangle();
	DoNoise();
	DoPCM();

	if (FSettings.soundq>0)
	{
		int32	*tmpo = &FCEUWaveHi[soundtsoffs];

		if (GameExpSound.HiFill)
			GameExpSound.HiFill();

		if ((x=timestamp)!=0) /* ==0 already checked, but timestamp can be volatile */
			do
			{
				uint32		b = *tmpo;

				*tmpo++ = (b&0xFFFF)+tpcmwlookup[b>>16&0xFF]+rectwlookup[b>>24];
			} while(--x!=0);
		end = NeoFilterSound(FCEUWaveHi,WaveFinal,SOUNDTS,&left);

		memmove(FCEUWaveHi,FCEUWaveHi+SOUNDTS-left,left*sizeof(uint32));
		memset(FCEUWaveHi+left,0,sizeof(FCEUWaveHi)-left*sizeof(uint32));

		if (GameExpSound.HiSync)
			GameExpSound.HiSync(left);

		ChannelBC[0] = ChannelBC[1] = ChannelBC[2] = ChannelBC[3] = ChannelBC[4] = left;
	} else
	{
		end = (SOUNDTS<<16)/soundtsinc;
		if (GameExpSound.Fill)
			GameExpSound.Fill(end&15);

		SexyFilter(FCEUWave,WaveFinal,end>>4);

		//if(FSettings.lowpass)
		// SexyFilter2(WaveFinal,end>>4);

		if (end&15) /* roll ending sample to beginning */
			FCEUWave[0] = FCEUWave[end>>4];
		FCEUWave[end>>4] = 0;
	}
nosoundo:
	if (FSettings.soundq>0)
		soundtsoffs = left;
	else
	{
		ChannelBC[0] = ChannelBC[1] = ChannelBC[2] = ChannelBC[3] = ChannelBC[4] = end&15;
		soundtsoffs = (soundtsinc*(end&15))>>16;
		end>>= 4;
	}
	inbuf = end;

	FCEU_WriteWaveData(WaveFinal,end); /* does nothing if sound recording is off */
	return end;
}

int GetSoundBuffer(int16 **W)
{
	*W = WaveFinal;
	return inbuf;
}

/* FIXME:  Find out what sound registers get reset on reset.  I know $4001/$4005 don't,
due to that whole MegaMan 2 Game Genie thing.
*/

void FCEUSND_Reset(void)
{
	IRQFrameMode = 0x0;
	Soundfhcnt = Soundfhinc;
	Soundfcnt = 0;

	nreg = 1;

	wlcount[0] = wlcount[1] = 2048;
	if (nesincsize) /* lq mode */
		sqacc[0] = sqacc[1] = ((uint32)2048<<17)/nesincsize;
	else
		sqacc[0] = sqacc[1] = 1;
	sweepon[0] = sweepon[1] = 0;
	curfreq[0] = curfreq[1] = 0;

	wlcount[2] = 1;	//2048;
	wlcount[3] = 2048;
	DMCHaveDMA = DMCHaveSample = 0;
	SIRQStat = 0x00;

	RawDALatch = 0x00;
	TriCount = 0;
	TriMode = 0;
	tristep = 0;
	EnabledChannels = 0;

	soundlencount[0] = soundlencount[1] = soundlencount[2] = soundlencount[3] = 0;

	DMCAddressLatch = 0;
	DMCSizeLatch = 0;
	DMCFormat = 0;
	DMCAddress = 0;
	DMCSize = 0;
	DMCShift = 0;
}

void FCEUSND_Power(void)
{
	int		x;

	SetNESSoundMap();
	memset(PSG,0,sizeof(PSG));
	FCEUSND_Reset();

	/* added (actual may be in Reset, but this gives an option); 2009.12.27 08:32 EST */
	DMCacc = 1;
	DMCBitCount = 0;

	memset(FCEUWave,0,sizeof(FCEUWave));
	memset(FCEUWaveHi,0,sizeof(FCEUWaveHi));
	memset(EnvUnits,0,sizeof(EnvUnits));

	x = 5;
	do
		ChannelBC[x-1] = 0;
	while(--x!=0);
	soundtsoffs = 0;
	LoadDMCPeriod(DMCFormat&15);
}

static void SndSetFuncs(uint8 bits)
{
	if (FSettings.soundq>0)/*>=1*/
	{
		DoSQ1 = (bits&1) ? RDoSQ1:Dummyfunc;
		DoSQ2 = (bits&2) ? RDoSQ2:Dummyfunc;
		DoTriangle = (bits&4) ? RDoTriangle:Dummyfunc;
		DoNoise = (bits&8) ? RDoNoise:Dummyfunc;
		DoPCM = (bits&0x10) ? RDoPCM:Dummyfunc;
	} else
	{
		DoSQ1 = (bits&1) ? RDoSQLQ:Dummyfunc;
		DoSQ2 = (bits&2) ? RDoSQLQ:Dummyfunc;
		DoTriangle = (bits&4) ? RDoTriangleNoisePCMLQ:Dummyfunc;
		DoNoise = (bits&8) ? RDoTriangleNoisePCMLQ:Dummyfunc;
		DoPCM = (bits&0x10) ? RDoTriangleNoisePCMLQ:Dummyfunc;
	}
}

/* Is the correct PAL sound chip exactly 0.8MHz?  The CPU is certainly not 1.6MHz. */
/* From what I recall, PAL emulation naturally comes close to the console notation */
/*and tempo; however, NTSC at 4 notes behind at rate, while 2 notes ahead in sound.*/
/*Playback seems to match exactly.  Certain cycle counts in ppu.c probably need to */
/*be adjusted for correct timing here (a late music loop is obvious in some carts).*/
/* The 2A07 PAL CPU apparently operates it's own internal sound chip at a slightly */
/*different rate than what's emulated here. */

#define ADJUSTED_PAL_CPU_RATIO		(double)(PAL_CPU/1600000)

void SetSoundVariables(void)
{
	uint64		ratio;
	uint32		Rate;
	int		x;

	Soundfhinc = PAL ? 16626*2:14915*2; // *2 CPU clock rate

	memcpy(CurrentNoiseFreqTable,PAL ? PALNoiseFreqTable:NTSCNoiseFreqTable,
		sizeof(CurrentNoiseFreqTable));
	memcpy(CurrentDMCTable,PAL ? PALDMCTable:NTSCDMCTable,
		sizeof(CurrentDMCTable));

	rectwlookup[0] = tpcmwlookup[0] = 0;

	if ((Rate=FSettings.SndRate*16)!=0) /* x16 for LQ anti-aliasing */
	{ /* build the wave amplitude LUTs */
		double	numer; /*16*16*16*4*/

		x = 1, numer = 95.52*(1L<<12+2-(FSettings.soundq ? 0+8:4+8))
			*(double)FSettings.soundrectamp;
		do
			rectwlookup[x] = numer/((double)8128/(double)x+100);
		while(++x<32);
		x = 1, numer = 163.67*(1L<<12+2-(FSettings.soundq ? 0:4));
		do
			tpcmwlookup[x] = numer/((double)24329/(double)x+100);
		while(++x<203);

		SndSetFuncs(FSettings.soundactivech);
	} else
	{
		SndSetFuncs(0);
		return;
	}

	MakeFilters(Rate>>4);

	if (GameExpSound.RChange)
		GameExpSound.RChange();

	if (PAL)
		_int64xset(ratio,(1L<<17)*(double)PAL_CPU);
	else
		_int64xset(ratio,(1L<<17)*(double)NTSC_CPU);

	nesincsize = uint64divu32(ratio,Rate);

	sqacc[0] = sqacc[1] = 0;
	memset(ChannelBC,0,sizeof(ChannelBC));

	LoadDMCPeriod(DMCFormat&15); /* For changing from PAL to NTSC */

	soundtsinc = (unsigned)nesincsize;

	if (PAL && ConsolePitchFix)
	{
		nesincsize = ufract32mul(nesincsize*2,4294967296.0/ADJUSTED_PAL_CPU_RATIO+.9)+1>>1;
		Soundfhinc = 16626.0*2/ADJUSTED_PAL_CPU_RATIO;
	}

	if (EmulationSpeedRatio!=1000000000UL && EmulationSpeedRatio!=0)
	{ /* adjust PCM tables for rate adjustment */
		x = 0;
		do
		{
			CurrentDMCTable[x] =
				umuldiv(CurrentDMCTable[x],EmulationSpeedRatio,500000000UL)+1>>1;
		} while(++x<sizeof(CurrentDMCTable)/sizeof(CurrentDMCTable[0]));

		Soundfhinc = umuldiv(Soundfhinc,1000000000UL,EmulationSpeedRatio);
		soundtsinc = umuldiv(soundtsinc,EmulationSpeedRatio,1000000000UL);
	}

	soundtsinc>>= 1;

	Soundfhinc = 24*(Soundfhinc+1>>1);
}

void FCEUI_Sound(int Rate)
{
	FSettings.SndRate = Rate;
	SetSoundVariables();
}

/* added; 2008.09.12 */
void FCEUI_SetSoundActiveChannels(uint8 bits)
{ /* flush before setting */
	DoSQ1();
	DoSQ2();
	DoTriangle();
	DoNoise();
	DoPCM();

	FSettings.soundactivech = bits;

	if (FSettings.SndRate!=0)
		SndSetFuncs(bits);
}

void FCEUI_SetLowPass(int q)
{
	FSettings.lowpass = q;
}

void FCEUI_SetSoundQuality(int quality)
{
	FSettings.soundq = quality;
	SetSoundVariables();
}

void FCEUI_SetSoundVolume(uint32 volume)
{
	FSettings.SoundVolume = volume;
}

/* added; 2010.03.23 */
void FCEUI_SetRectFilterAmp(int amplitude)
{
	FSettings.soundrectamp = amplitude;
	SetSoundVariables();
}

SFORMAT FCEUSND_STATEINFO[] =
{
	{ &Soundfhcnt,4|FCEUSTATE_RLSB,"FHCN"},
	{ &Soundfcnt,1,"FCNT"},
	{ PSG,0x10,"PSG"},
	{ &EnabledChannels,1,"ENCH"},
	{ &IRQFrameMode,1,"IQFM"},
	{ &nreg,2|FCEUSTATE_RLSB,"NREG"},
	{ &TriMode,1,"TRIM"},
	{ &TriCount,1,"TRIC"},

	{ &EnvUnits[0].Speed,1,"E0SP"},
	{ &EnvUnits[1].Speed,1,"E1SP"},
	{ &EnvUnits[2].Speed,1,"E2SP"},

	{ &EnvUnits[0].Mode,1,"E0MO"},
	{ &EnvUnits[1].Mode,1,"E1MO"},
	{ &EnvUnits[2].Mode,1,"E2MO"},

	{ &EnvUnits[0].DecCountTo1,1,"E0D1"},
	{ &EnvUnits[1].DecCountTo1,1,"E1D1"},
	{ &EnvUnits[2].DecCountTo1,1,"E2D1"},

	{ &EnvUnits[0].decvolume,1,"E0DV"},
	{ &EnvUnits[1].decvolume,1,"E1DV"},
	{ &EnvUnits[2].decvolume,1,"E2DV"},

	{ &soundlencount[0],4|FCEUSTATE_RLSB,"LEN0"},
	{ &soundlencount[1],4|FCEUSTATE_RLSB,"LEN1"},
	{ &soundlencount[2],4|FCEUSTATE_RLSB,"LEN2"},
	{ &soundlencount[3],4|FCEUSTATE_RLSB,"LEN3"},
	{ sweepon,2,"SWEE"},
	{ &curfreq[0],4|FCEUSTATE_RLSB,"CRF1"},
	{ &curfreq[1],4|FCEUSTATE_RLSB,"CRF2"},
	{ SweepCount,2,"SWCT"},

	{ &SIRQStat,1,"SIRQ"},

	{ &DMCacc,4|FCEUSTATE_RLSB,"5ACC"},
	{ &DMCBitCount,1,"5BIT"},
	{ &DMCAddress,4|FCEUSTATE_RLSB,"5ADD"},
	{ &DMCSize,4|FCEUSTATE_RLSB,"5SIZ"},
	{ &DMCShiftF,2,"5SHF"},

	{ &DMCHaveDMA,1,"5HVD"}, //note five chars... truncated; 5HVDM, 5HVSP
	{ &DMCHaveSample,1,"5HVS"},

	{ &DMCSizeLatch,1,"5SZL"},
	{ &DMCAddressLatch,1,"5ADL"},
	{ &DMCFormat,1,"5FMT"},
	{ &RawDALatch,1,"RWDA"},
	{ 0}
};

void FCEUSND_SaveState()
{
}

void FCEUSND_LoadState(int v)
{
(void)v;
	LoadDMCPeriod(DMCFormat&15);
	DMCAddress&= 0x7FFF;
	RawDALatch&= 0x7F;
	DMCShiftF[0] = 0; //for an assmebly optimization
}
